@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Create Version
  * 
@@ -36,8 +36,6 @@
 #include "cvs.h"
 
 static char *findslash PROTO((char *start, char *p));
-static int build_dirs_and_chdir PROTO((char *dir, char *prepath, char *realdir,
-				 int sticky));
 static int checkout_proc PROTO((int *pargc, char **argv, char *where,
 		          char *mwhere, char *mfile, int shorten,
 		          int local_specified, char *omodule,
@@ -46,10 +44,12 @@ static int safe_location PROTO((void));
 
 static const char *const checkout_usage[] =
 {
-    "Usage:\n  %s %s [-ANPcflnps] [-r rev | -D date] [-d dir] [-k kopt] modules...\n",
+    "Usage:\n  %s %s [-ANPRcflnps] [-r rev | -D date] [-d dir]\n",
+    "    [-j rev1] [-j rev2] [-k kopt] modules...\n",
     "\t-A\tReset any sticky tags/date/kopts.\n",
     "\t-N\tDon't shorten module paths if -d specified.\n",
     "\t-P\tPrune empty directories.\n",
+    "\t-R\tProcess directories recursively.\n",
     "\t-c\t\"cat\" the module database.\n",
     "\t-f\tForce a head revision match if tag/date not found.\n",
     "\t-l\tLocal directory only, not recursive\n",
@@ -61,20 +61,23 @@ static const char *const checkout_usage[] =
     "\t-d dir\tCheck out into dir instead of module name.\n",
     "\t-k kopt\tUse RCS kopt -k option on checkout.\n",
     "\t-j rev\tMerge in changes made between current revision and rev.\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
 static const char *const export_usage[] =
 {
-    "Usage: %s %s [-NPfln] [-r rev | -D date] [-d dir] [-k kopt] module...\n",
+    "Usage: %s %s [-NRfln] [-r rev | -D date] [-d dir] [-k kopt] module...\n",
     "\t-N\tDon't shorten module paths if -d specified.\n",
     "\t-f\tForce a head revision match if tag/date not found.\n",
     "\t-l\tLocal directory only, not recursive\n",
+    "\t-R\tProcess directories recursively (default).\n",
     "\t-n\tDo not run module program (if any).\n",
-    "\t-r rev\tCheck out revision or tag.\n",
-    "\t-D date\tCheck out revisions as of date.\n",
-    "\t-d dir\tCheck out into dir instead of module name.\n",
+    "\t-r rev\tExport revision or tag.\n",
+    "\t-D date\tExport revisions as of date.\n",
+    "\t-d dir\tExport into dir instead of module name.\n",
     "\t-k kopt\tUse RCS kopt -k option on checkout.\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
@@ -118,13 +121,13 @@ checkout (argc, argv)
     if (strcmp (command_name, "export") == 0)
     {
         m_type = EXPORT;
-	valid_options = "Nnk:d:flRQqr:D:";
+	valid_options = "+Nnk:d:flRQqr:D:";
 	valid_usage = export_usage;
     }
     else
     {
         m_type = CHECKOUT;
-	valid_options = "ANnk:d:flRpQqcsr:D:j:P";
+	valid_options = "+ANnk:d:flRpQqcsr:D:j:P";
 	valid_usage = checkout_usage;
     }
 
@@ -134,7 +137,7 @@ checkout (argc, argv)
     ign_setup ();
     wrap_setup ();
 
-    optind = 1;
+    optind = 0;
     while ((c = getopt (argc, argv, valid_options)) != -1)
     {
 	switch (c)
@@ -220,8 +223,11 @@ checkout (argc, argv)
     if (shorten == -1)
 	shorten = 0;
 
-    if ((!(cat + status) && argc == 0) || ((cat + status) && argc != 0))
-	usage (valid_usage);
+    if ((cat || status) && argc != 0)
+	error (1, 0, "-c and -s must not get any arguments");
+
+    if (!(cat || status) && argc == 0)
+	error (1, 0, "must specify at least one module or directory");
 
     if (where && pipeout)
 	error (1, 0, "-d and -p are mutually exclusive");
@@ -229,20 +235,10 @@ checkout (argc, argv)
     if (strcmp (command_name, "export") == 0)
     {
 	if (!tag && !date)
-	{
-	    error (0, 0, "must specify a tag or date");
-	    usage (valid_usage);
-	}
+	    error (1, 0, "must specify a tag or date");
+
 	if (tag && isdigit (tag[0]))
 	    error (1, 0, "tag `%s' must be a symbolic tag", tag);
-/*
- * mhy 950615: -kv doesn't work for binaries with RCS keywords.
- * Instead use the default provided in the RCS file (-ko for binaries).
- */
-#if 0
-	if (!options)
-	  options = RCS_check_kflag ("v");/* -kv is default */
-#endif
     }
 
     if (!safe_location()) {
@@ -263,7 +259,7 @@ checkout (argc, argv)
            server send the check-in and update programs for the
            various modules/dirs requested.  If we turn this off and
            simply request the names of the modules and directories (as
-           below in !expand_modules), those files (CVS/Checking.prog
+           below in !expand_modules), those files (CVS/Checkin.prog
            or CVS/Update.prog) don't get created.  Grrr.  */
 	
 	expand_modules = (!cat && !status && !pipeout
@@ -278,10 +274,14 @@ checkout (argc, argv)
 	    client_expand_modules (argc, argv, local);
 	}
 
-	if (!run_module_prog) send_arg ("-n");
-	if (local) send_arg ("-l");
-	if (pipeout) send_arg ("-p");
-	if (!force_tag_match) send_arg ("-f");
+	if (!run_module_prog)
+	    send_arg ("-n");
+	if (local)
+	    send_arg ("-l");
+	if (pipeout)
+	    send_arg ("-p");
+	if (!force_tag_match)
+	    send_arg ("-f");
 	if (aflag)
 	    send_arg("-A");
 	if (!shorten)
@@ -292,14 +292,10 @@ checkout (argc, argv)
 	if (cat)
 	    send_arg("-c");
 	if (where != NULL)
-	{
 	    option_with_arg ("-d", where);
-	}
 	if (status)
 	    send_arg("-s");
-	if (strcmp (command_name, "export") != 0
-	    && options != NULL
-	    && options[0] != '\0')
+	if (options != NULL && options[0] != '\0')
 	    send_arg (options);
 	option_with_arg ("-r", tag);
 	if (date)
@@ -311,7 +307,7 @@ checkout (argc, argv)
 
 	if (expand_modules)
 	{
-	    client_send_expansions (local, where);
+	    client_send_expansions (local, where, 1);
 	}
 	else
 	{
@@ -332,56 +328,22 @@ checkout (argc, argv)
     if (cat || status)
     {
 	cat_module (status);
+	if (options)
+	    free (options);
 	return (0);
     }
     db = open_module ();
 
-    /*
-     * if we have more than one argument and where was specified, we make the
-     * where, cd into it, and try to shorten names as much as possible.
-     * Otherwise, we pass the where as a single argument to do_module.
-     */
-    if (argc > 1 && where != NULL)
-    {
-	char repository[PATH_MAX];
 
-	(void) CVS_MKDIR (where, 0777);
-	if ( CVS_CHDIR (where) < 0)
-	    error (1, errno, "cannot chdir to %s", where);
-	preload_update_dir = xstrdup (where);
-	where = (char *) NULL;
-	if (!isfile (CVSADM))
-	{
-	    (void) sprintf (repository, "%s/%s/%s", CVSroot_directory,
-			    CVSROOTADM, CVSNULLREPOS);
-	    if (!isfile (repository))
-	    {
-		mode_t omask;
-		omask = umask (cvsumask);
-		(void) CVS_MKDIR (repository, 0777);
-		(void) umask (omask);
-	    }
+    /* If we've specified something like "cvs co foo/bar baz/quux"
+       don't try to shorten names.  There are a few cases in which we
+       could shorten (e.g. "cvs co foo/bar foo/baz"), but we don't
+       handle those yet.  Better to have an extra directory created
+       than the thing checked out under the wrong directory name. */
 
-	    /* I'm not sure whether this check is redundant.  */
-	    if (!isdir (repository))
-		error (1, 0, "there is no repository %s", repository);
+    if (argc > 1)
+	shorten = 0;
 
-	    Create_Admin (".", preload_update_dir, repository,
-			  (char *) NULL, (char *) NULL);
-	    if (!noexec)
-	    {
-		FILE *fp;
-
-		fp = open_file (CVSADM_ENTSTAT, "w+");
-		if (fclose(fp) == EOF)
-		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
-#ifdef SERVER_SUPPORT
-		if (server_active)
-		    server_set_entstat (preload_update_dir, repository);
-#endif
-	    }
-	}
-    }
 
     /* If we will be calling history_write, work out the name to pass
        it.  */
@@ -398,47 +360,34 @@ checkout (argc, argv)
 	    history_name = date;
     }
 
-    /*
-     * if where was specified (-d) and we have not taken care of it already
-     * with the multiple arg stuff, and it was not a simple directory name
-     * but rather a path, we strip off everything but the last component and
-     * attempt to cd to the indicated place.  where then becomes simply the
-     * last component
-     */
-    if (where != NULL && strchr (where, '/') != NULL)
-    {
-	char *slash;
-
-	slash = strrchr (where, '/');
-	*slash = '\0';
-
-	if ( CVS_CHDIR (where) < 0)
-	    error (1, errno, "cannot chdir to %s", where);
-
-	preload_update_dir = xstrdup (where);
-
-	where = slash + 1;
-	if (*where == '\0')
-	    where = NULL;
-    }
 
     for (i = 0; i < argc; i++)
 	err += do_module (db, argv[i], m_type, "Updating", checkout_proc,
 			  where, shorten, local, run_module_prog,
 			  (char *) NULL);
     close_module (db);
+    if (options)
+	free (options);
     return (err);
 }
 
 static int
 safe_location ()
 {
-    char current[PATH_MAX];
+    char *current;
     char hardpath[PATH_MAX+5];
     size_t hardpath_len;
     int  x;
+    int retval;
 
+#ifdef HAVE_READLINK
+    /* FIXME-arbitrary limit: should be retrying this like xgetwd.
+       But how does readlink let us know that the buffer was too small?
+       (by returning sizeof hardpath - 1?).  */
     x = readlink(CVSroot_directory, hardpath, sizeof hardpath - 1);
+#else
+    x = -1;
+#endif
     if (x == -1)
     {
         strcpy(hardpath, CVSroot_directory);
@@ -447,22 +396,83 @@ safe_location ()
     {
         hardpath[x] = '\0';
     }
-    getwd (current);
+    current = xgetwd ();
+    if (current == NULL)
+	error (1, errno, "could not get working directory");
     hardpath_len = strlen (hardpath);
-    if (strncmp (current, hardpath, hardpath_len) == 0)
+    if (strlen (current) >= hardpath_len
+	&& strncmp (current, hardpath, hardpath_len) == 0)
     {
 	if (/* Current is a subdirectory of hardpath.  */
 	    current[hardpath_len] == '/'
 
 	    /* Current is hardpath itself.  */
 	    || current[hardpath_len] == '\0')
-	    return 0;
+	    retval = 0;
 	else
 	    /* It isn't a problem.  For example, current is
 	       "/foo/cvsroot-bar" and hardpath is "/foo/cvsroot".  */
-	    return 1;
+	    retval = 1;
+    }
+    else
+	retval = 1;
+    free (current);
+    return retval;
+}
+
+struct dir_to_build
+{
+    /* What to put in CVS/Repository.  */
+    char *repository;
+    /* The path to the directory.  */
+    char *dirpath;
+
+    struct dir_to_build *next;
+};
+
+static int build_dirs_and_chdir PROTO ((struct dir_to_build *list,
+					int sticky));
+
+static void build_one_dir PROTO ((char *, char *, int));
+
+static void
+build_one_dir (repository, dirpath, sticky)
+    char *repository;
+    char *dirpath;
+    int sticky;
+{
+    FILE *fp;
+
+    if (!isfile (CVSADM) && strcmp (command_name, "export") != 0)
+    {
+	/* I suspect that this check could be omitted.  */
+	if (!isdir (repository))
+	    error (1, 0, "there is no repository %s", repository);
+
+	if (Create_Admin (".", dirpath, repository,
+			  sticky ? (char *) NULL : tag,
+			  sticky ? (char *) NULL : date,
+
+			  /* FIXME?  This is a guess.  If it is important
+			     for nonbranch to be set correctly here I
+			     think we need to write it one way now and
+			     then rewrite it later via WriteTag, once
+			     we've had a chance to call RCS_nodeisbranch
+			     on each file.  */
+			  0, 1))
+	    return;
+
+	if (!noexec)
+	{
+	    fp = open_file (CVSADM_ENTSTAT, "w+");
+	    if (fclose (fp) == EOF)
+		error (1, errno, "cannot close %s", CVSADM_ENTSTAT);
+#ifdef SERVER_SUPPORT
+	    if (server_active)
+		server_set_entstat (dirpath, repository);
+#endif
+	}
     }
-    return (1);
 }
 
 /*
@@ -470,11 +480,11 @@ safe_location ()
  */
 /* ARGSUSED */
 static int
-checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
+checkout_proc (pargc, argv, where_orig, mwhere, mfile, shorten,
 	       local_specified, omodule, msg)
     int *pargc;
     char **argv;
-    char *where;
+    char *where_orig;
     char *mwhere;
     char *mfile;
     int shorten;
@@ -485,12 +495,9 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
     int err = 0;
     int which;
     char *cp;
-    char *cp2;
-    char repository[PATH_MAX];
-    char xwhere[PATH_MAX];
+    char *repository;
     char *oldupdate = NULL;
-    char *prepath;
-    char *realdirs;
+    char *where;
 
     /*
      * OK, so we're doing the checkout! Our args are as follows: 
@@ -498,139 +505,181 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
      *  where contains where to put it (if supplied by checkout) 
      *  mwhere contains the module name or -d from module file 
      *  mfile says do only that part of the module
-     *  shorten = TRUE says shorten as much as possible 
+     *  shorten = 1 says shorten as much as possible 
      *  omodule is the original arg to do_module()
      */
 
-    /* set up the repository (maybe) for the bottom directory */
+    /* Set up the repository (maybe) for the bottom directory.
+       Allocate more space than we need so we don't need to keep
+       reallocating this string. */
+    repository = xmalloc (strlen (CVSroot_directory)
+			  + strlen (argv[0])
+			  + (mfile == NULL ? 0 : strlen (mfile))
+			  + 10);
     (void) sprintf (repository, "%s/%s", CVSroot_directory, argv[0]);
+    Sanitize_Repository_Name (repository);
+
 
     /* save the original value of preload_update_dir */
     if (preload_update_dir != NULL)
 	oldupdate = xstrdup (preload_update_dir);
 
-    /* fix up argv[] for the case of partial modules */
+
+    /* Allocate space and set up the where variable.  We allocate more
+       space than necessary here so that we don't have to keep
+       reallocaing it later on. */
+    
+    where = xmalloc (strlen (argv[0])
+		     + (mfile == NULL ? 0 : strlen (mfile))
+		     + (mwhere == NULL ? 0 : strlen (mwhere))
+		     + (where_orig == NULL ? 0 : strlen (where_orig))
+		     + 10);
+
+    /* Yes, this could be written in a less verbose way, but in this
+       form it is quite easy to read.
+    
+       FIXME?  The following code that sets should probably be moved
+       to do_module in modules.c, since there is similar code in
+       patch.c and rtag.c. */
+    
+    if (shorten)
+    {
+	if (where_orig != NULL)
+	{
+	    /* If the user has specified a directory with `-d' on the
+	       command line, use it preferentially, even over the `-d'
+	       flag in the modules file. */
+    
+	    (void) strcpy (where, where_orig);
+	}
+	else if (mwhere != NULL)
+	{
+	    /* Second preference is the value of mwhere, which is from
+	       the `-d' flag in the modules file. */
+
+	    (void) strcpy (where, mwhere);
+	}
+	else
+	{
+	    /* Third preference is the directory specified in argv[0]
+	       which is this module'e directory in the repository. */
+	    
+	    (void) strcpy (where, argv[0]);
+	}
+    }
+    else
+    {
+	/* Use the same preferences here, bug don't shorten -- that
+           is, tack on where_orig if it exists. */
+
+	*where = '\0';
+
+	if (where_orig != NULL)
+	{
+	    (void) strcat (where, where_orig);
+	    (void) strcat (where, "/");
+	}
+
+	if (mwhere != NULL)
+	    (void) strcat (where, mwhere);
+	else
+	    (void) strcat (where, argv[0]);
+    }
+    strip_trailing_slashes (where); /* necessary? */
+
+
+    /* At this point, the user may have asked for a single file or
+       directory from within a module.  In that case, we should modify
+       where, repository, and argv as appropriate. */
+
     if (mfile != NULL)
     {
-	char file[PATH_MAX];
+	/* The mfile variable can have one or more path elements.  If
+	   it has multiple elements, we want to tack those onto both
+	   repository and where.  The last element may refer to either
+	   a file or directory.  Here's what to do:
 
-	/* if mfile is really a path, straighten it out first */
-	if ((cp = strrchr (mfile, '/')) != NULL)
+	   it refers to a directory
+	     -> simply tack it on to where and repository
+	   it refers to a file
+	     -> munge argv to contain `basename mfile` */
+
+	char *cp;
+	char *path;
+
+
+	/* Paranoia check. */
+
+	if (mfile[strlen (mfile) - 1] == '/')
 	{
-	    *cp = 0;
+	    error (0, 0, "checkout_proc: trailing slash on mfile (%s)!",
+		   mfile);
+	}
+
+
+	/* Does mfile have multiple path elements? */
+
+	cp = strrchr (mfile, '/');
+	if (cp != NULL)
+	{
+	    *cp = '\0';
 	    (void) strcat (repository, "/");
 	    (void) strcat (repository, mfile);
-
-	    /*
-	     * Now we need to fill in the where correctly. if !shorten, tack
-	     * the rest of the path onto where if where is filled in
-	     * otherwise tack the rest of the path onto mwhere and make that
-	     * the where
-	     * 
-	     * If shorten is enabled, we might use mwhere to set where if 
-	     * nobody set it yet, so we'll need to setup mwhere as the last
-	     * component of the path we are tacking onto repository
-	     */
-	    if (!shorten)
-	    {
-		if (where != NULL)
-		    (void) sprintf (xwhere, "%s/%s", where, mfile);
-		else
-		    (void) sprintf (xwhere, "%s/%s", mwhere, mfile);
-		where = xwhere;
-	    }
-	    else
-	    {
-		char *slash;
-
-		if ((slash = strrchr (mfile, '/')) != NULL)
-		    mwhere = slash + 1;
-		else
-		    mwhere = mfile;
-	    }
+	    (void) strcat (where, "/");
+	    (void) strcat (where, mfile);
 	    mfile = cp + 1;
 	}
+	
 
-	(void) sprintf (file, "%s/%s", repository, mfile);
-	if (isdir (file))
+	/* Now mfile is a single path element. */
+
+	path = xmalloc (strlen (repository) + strlen (mfile) + 5);
+	(void) sprintf (path, "%s/%s", repository, mfile);
+	if (isdir (path))
 	{
+	    /* It's a directory, so tack it on to repository and
+               where, as we did above. */
 
-	    /*
-	     * The portion of a module was a directory, so kludge up where to
-	     * be the subdir, and fix up repository
-	     */
-	    (void) strcpy (repository, file);
-
-	    /*
-	     * At this point, if shorten is not enabled, we make where either
-	     * where with mfile concatenated, or if where hadn't been set we
-	     * set it to mwhere with mfile concatenated.
-	     * 
-	     * If shorten is enabled and where hasn't been set yet, then where
-	     * becomes mfile
-	     */
-	    if (!shorten)
-	    {
-		if (where != NULL)
-		    (void) sprintf (xwhere, "%s/%s", where, mfile);
-		else
-		    (void) sprintf (xwhere, "%s/%s", mwhere, mfile);
-		where = xwhere;
-	    }
-	    else if (where == NULL)
-		where = mfile;
+	    (void) strcat (repository, "/");
+	    (void) strcat (repository, mfile);
+	    (void) strcat (where, "/");
+	    (void) strcat (where, mfile);
 	}
 	else
 	{
+	    /* It's a file, which means we have to screw around with
+               argv. */
+
 	    int i;
 
-	    /*
-	     * The portion of a module was a file, so kludge up argv to be
-	     * correct
-	     */
-	    for (i = 1; i < *pargc; i++)/* free the old ones */
+
+	    /* Paranoia check. */
+	    
+	    if (*pargc > 1)
+	    {
+		error (0, 0, "checkout_proc: trashing argv elements!");
+		for (i = 1; i < *pargc; i++)
+		{
+		    error (0, 0, "checkout_proc: argv[%d] `%s'",
+			   i, argv[i]);
+		}
+	    }
+
+	    for (i = 1; i < *pargc; i++)
 		free (argv[i]);
-	    argv[1] = xstrdup (mfile);	/* set up the new one */
-	    *pargc = 2;
-
-	    /* where gets mwhere if where isn't set */
-	    if (where == NULL)
-		where = mwhere;
-	}
-    }
-
-    /*
-     * if shorten is enabled and where isn't specified yet, we pluck the last
-     * directory component of argv[0] and make it the where
-     */
-    if (shorten && where == NULL)
-    {
-	if ((cp = strrchr (argv[0], '/')) != NULL)
-	{
-	    (void) strcpy (xwhere, cp + 1);
-	    where = xwhere;
-	}
-    }
-
-    /* if where is still NULL, use mwhere if set or the argv[0] dir */
-    if (where == NULL)
-    {
-	if (mwhere)
-	    where = mwhere;
-	else
-	{
-	    (void) strcpy (xwhere, argv[0]);
-	    where = xwhere;
+	    argv[1] = xstrdup (mfile);
+	    (*pargc) = 2;
 	}
+	free (path);
     }
 
     if (preload_update_dir != NULL)
     {
-	char tmp[PATH_MAX];
-
-	(void) sprintf (tmp, "%s/%s", preload_update_dir, where);
-	free (preload_update_dir);
-	preload_update_dir = xstrdup (tmp);
+	preload_update_dir =
+	    xrealloc (preload_update_dir,
+		      strlen (preload_update_dir) + strlen (where) + 5);
+	strcat (preload_update_dir, "/");
+	strcat (preload_update_dir, where);
     }
     else
 	preload_update_dir = xstrdup (where);
@@ -638,6 +687,12 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
     /*
      * At this point, where is the directory we want to build, repository is
      * the repository for the lowest level of the path.
+     *
+     * We need to tell build_dirs not only the path we want it to
+     * build, but also the repositories we want it to populate the
+     * path with.  To accomplish this, we walk the path backwards, one
+     * pathname component at a time, constucting a linked list of
+     * struct dir_to_build.
      */
 
     /*
@@ -646,24 +701,131 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
      */
     if (!pipeout)
     {
+	struct dir_to_build *head;
+	char *reposcopy;
 
-	/*
-	 * We need to tell build_dirs not only the path we want it to build,
-	 * but also the repositories we want it to populate the path with. To
-	 * accomplish this, we pass build_dirs a ``real path'' with valid
-	 * repositories and a string to pre-pend based on how many path
-	 * elements exist in where. Big Black Magic
-	 */
-	prepath = xstrdup (repository);
+	if (strncmp (repository, CVSroot_directory,
+		     strlen (CVSroot_directory)) != 0)
+	    error (1, 0, "\
+internal error: %s doesn't start with %s in checkout_proc",
+		   repository, CVSroot_directory);
+
+	/* We always create at least one directory, which corresponds to
+	   the entire strings for WHERE and REPOSITORY.  */
+	head = (struct dir_to_build *) xmalloc (sizeof (struct dir_to_build));
+	/* Special marker to indicate that we don't want build_dirs_and_chdir
+	   to create the CVSADM directory for us.  */
+	head->repository = NULL;
+	head->dirpath = xstrdup (where);
+	head->next = NULL;
+
+
+	/* Make a copy of the repository name to play with. */
+	reposcopy = xstrdup (repository);
+
+	/* FIXME: this should be written in terms of last_component instead
+	   of hardcoding '/'.  This presumably affects OS/2, NT, &c, if
+	   the user specifies '\'.  Likewise for the call to findslash.  */
 	cp = strrchr (where, '/');
-	cp2 = strrchr (prepath, '/');
 	while (cp != NULL)
 	{
+	    struct dir_to_build *new;
+	    new = (struct dir_to_build *)
+		xmalloc (sizeof (struct dir_to_build));
+	    new->dirpath = xmalloc (strlen (where));
+	    strncpy (new->dirpath, where, cp - where);
+	    new->dirpath[cp - where] = '\0';
+
+	    /* Now figure out what repository directory to generate.
+               The most complete case would be something like this:
+
+	       The modules file contains
+	         foo -d bar/baz quux
+
+	       The command issued was:
+	         cvs co -d what/ever -N foo
+	       
+	       The results in the CVS/Repository files should be:
+	         .     -> .          (this is where we executed the cmd)
+		 what  -> Emptydir   (generated dir -- not in repos)
+		 ever  -> .          (same as "cd what/ever; cvs co -N foo")
+		 bar   -> Emptydir   (generated dir -- not in repos)
+		 baz   -> quux       (finally!) */
+	    
+	    if (strcmp (reposcopy, CVSroot_directory) == 0)
+	    {
+		/* We can't walk up past CVSROOT.  Instead, the
+                   repository should be Emptydir. */
+		new->repository = emptydir_name ();
+	    }
+	    else
+	    {
+		if ((where_orig != NULL)
+		    && (strcmp (new->dirpath, where_orig) == 0))
+		{
+		    /* It's the case that the user specified a
+		     * destination directory with the "-d" flag.  The
+		     * repository in this directory should be "."
+		     * since the user's command is equivalent to:
+		     *
+		     *   cd <dir>; cvs co blah   */
+
+		    strcpy (reposcopy, CVSroot_directory);
+		    goto allocate_repos;
+		}
+		else if (mwhere != NULL)
+		{
+		    /* This is a generated directory, so point to
+                       CVSNULLREPOS. */
+
+		    new->repository = emptydir_name ();
+		}
+		else
+		{
+		    /* It's a directory in the repository! */
+		    
+		    char *rp = strrchr (reposcopy, '/');
+		    
+		    /* We'll always be below CVSROOT, but check for
+		       paranoia's sake. */
+		    if (rp == NULL)
+			error (1, 0,
+			       "internal error: %s doesn't contain a slash",
+			       reposcopy);
+			   
+		    *rp = '\0';
+		
+		allocate_repos:
+		    new->repository = xmalloc (strlen (reposcopy) + 5);
+		    (void) strcpy (new->repository, reposcopy);
+		    
+		    if (strcmp (reposcopy, CVSroot_directory) == 0)
+		    {
+			/* Special case -- the repository name needs
+			   to be "/path/to/repos/." (the trailing dot
+			   is important).  We might be able to get rid
+			   of this after the we check out the other
+			   code that handles repository names. */
+			(void) strcat (new->repository, "/.");
+		    }
+		}
+	    }
+	    
+	    new->next = head;
+	    head = new;
+
 	    cp = findslash (where, cp - 1);
-	    cp2 = findslash (prepath, cp2 - 1);
 	}
-	*cp2 = '\0';
-	realdirs = cp2 + 1;
+
+	/* clean up */
+	free (reposcopy);
+
+	/* The top-level CVSADM directory should always be
+           CVSroot_directory.  Create it.
+	
+	   It may be argued that we shouldn't set any sticky bits for
+	   the top-level repository.  FIXME?  */
+	build_one_dir (CVSroot_directory, ".", *pargc <= 1);
 
 	/*
 	 * build dirs on the path if necessary and leave us in the bottom
@@ -671,18 +833,13 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	 * subdir yet, but all the others contain CVS and Entries.Static
 	 * files
 	 */
-	if (build_dirs_and_chdir (where, prepath, realdirs, *pargc <= 1) != 0)
+	if (build_dirs_and_chdir (head, *pargc <= 1) != 0)
 	{
 	    error (0, 0, "ignoring module %s", omodule);
-	    free (prepath);
-	    free (preload_update_dir);
-	    preload_update_dir = oldupdate;
-	    return (1);
+	    err = 1;
+	    goto out;
 	}
 
-	/* clean up */
-	free (prepath);
-
 	/* set up the repository (or make sure the old one matches) */
 	if (!isfile (CVSADM))
 	{
@@ -695,7 +852,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 		    error (1, 0, "there is no repository %s", repository);
 
 		Create_Admin (".", preload_update_dir, repository,
-			      (char *) NULL, (char *) NULL);
+			      (char *) NULL, (char *) NULL, 0, 0);
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
@@ -710,7 +867,15 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 		if (!isdir (repository))
 		    error (1, 0, "there is no repository %s", repository);
 
-		Create_Admin (".", preload_update_dir, repository, tag, date);
+		Create_Admin (".", preload_update_dir, repository, tag, date,
+
+			      /* FIXME?  This is a guess.  If it is important
+				 for nonbranch to be set correctly here I
+				 think we need to write it one way now and
+				 then rewrite it later via WriteTag, once
+				 we've had a chance to call RCS_nodeisbranch
+				 on each file.  */
+			      0, 0);
 	    }
 	}
 	else
@@ -725,9 +890,8 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 		       repos, repository);
 		error (0, 0, "ignoring module %s", omodule);
 		free (repos);
-		free (preload_update_dir);
-		preload_update_dir = oldupdate;
-		return (1);
+		err = 1;
+		goto out;
 	    }
 	    free (repos);
 	}
@@ -743,9 +907,8 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	if ( CVS_CHDIR (repository) < 0)
 	{
 	    error (0, errno, "cannot chdir to %s", repository);
-	    free (preload_update_dir);
-	    preload_update_dir = oldupdate;
-	    return (1);
+	    err = 1;
+	    goto out;
 	}
 	which = W_REPOS;
 	if (tag != NULL && !tag_validated)
@@ -789,14 +952,15 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 	if (strcmp (command_name, "export") != 0 && !pipeout)
 	    history_write ('O', preload_update_dir, history_name, where,
 			   repository);
+	else if (strcmp (command_name, "export") == 0 && !pipeout)
+	    history_write ('E', preload_update_dir, tag ? tag : date, where,
+			   repository);
 	err += do_update (0, (char **) NULL, options, tag, date,
 			  force_tag_match, 0 /* !local */ ,
 			  1 /* update -d */ , aflag, checkout_prune_dirs,
 			  pipeout, which, join_rev1, join_rev2,
 			  preload_update_dir);
-	free (preload_update_dir);
-	preload_update_dir = oldupdate;
-	return (err);
+	goto out;
     }
 
     if (!pipeout)
@@ -836,6 +1000,7 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 		free (line);
 	    }
 	    freevers_ts (&vers);
+	    freercsnode (&finfo.rcs);
 	}
 
 	Entries_Close (entries);
@@ -851,8 +1016,11 @@ checkout_proc (pargc, argv, where, mwhere, mfile, shorten,
 		      force_tag_match, local_specified, 1 /* update -d */,
 		      aflag, checkout_prune_dirs, pipeout, which, join_rev1,
 		      join_rev2, preload_update_dir);
+out:
     free (preload_update_dir);
     preload_update_dir = oldupdate;
+    free (where);
+    free (repository);
     return (err);
 }
 
@@ -863,75 +1031,75 @@ findslash (start, p)
 {
     while (p >= start && *p != '/')
 	p--;
+    /* FIXME: indexing off the start of the array like this is *NOT*
+       OK according to ANSI, and will break some of the time on certain
+       segmented architectures.  */
     if (p < start)
 	return (NULL);
     else
 	return (p);
 }
 
-/*
- * build all the dirs along the path to dir with CVS subdirs with appropriate
- * repositories and Entries.Static files
- */
+/* Return a newly malloc'd string containing a pathname for CVSNULLREPOS,
+   and make sure that it exists.  If there is an error creating the
+   directory, give a fatal error.  Otherwise, the directory is guaranteed
+   to exist when we return.  */
+char *
+emptydir_name ()
+{
+    char *repository;
+
+    repository = xmalloc (strlen (CVSroot_directory) 
+			  + sizeof (CVSROOTADM)
+			  + sizeof (CVSNULLREPOS)
+			  + 10);
+    (void) sprintf (repository, "%s/%s/%s", CVSroot_directory,
+		    CVSROOTADM, CVSNULLREPOS);
+    if (!isfile (repository))
+    {
+	mode_t omask;
+	omask = umask (cvsumask);
+	if (CVS_MKDIR (repository, 0777) < 0)
+	    error (1, errno, "cannot create %s", repository);
+	(void) umask (omask);
+    }
+    return repository;
+}
+
+/* Build all the dirs along the path to DIRS with CVS subdirs with appropriate
+   repositories.  If ->repository is NULL, do not create a CVSADM directory
+   for that subdirectory; just CVS_CHDIR into it.  */
 static int
-build_dirs_and_chdir (dir, prepath, realdir, sticky)
-    char *dir;
-    char *prepath;
-    char *realdir;
+build_dirs_and_chdir (dirs, sticky)
+    struct dir_to_build *dirs;
     int sticky;
 {
-    FILE *fp;
-    char repository[PATH_MAX];
-    char path[PATH_MAX];
-    char path2[PATH_MAX];
-    char *slash;
-    char *slash2;
-    char *cp;
-    char *cp2;
+    int retval = 0;
+    struct dir_to_build *nextdir;
 
-    (void) strcpy (path, dir);
-    (void) strcpy (path2, realdir);
-    for (cp = path, cp2 = path2;
-    (slash = strchr (cp, '/')) != NULL && (slash2 = strchr (cp2, '/')) != NULL;
-	 cp = slash + 1, cp2 = slash2 + 1)
+    while (dirs != NULL)
     {
-	*slash = '\0';
-	*slash2 = '\0';
-	Subdir_Register ((List *) NULL, (char *) NULL, cp);
-	(void) CVS_MKDIR (cp, 0777);
-	if ( CVS_CHDIR (cp) < 0)
+	char *dir = last_component (dirs->dirpath);
+
+	mkdir_if_needed (dir);
+	Subdir_Register (NULL, NULL, dir);
+	if (CVS_CHDIR (dir) < 0)
 	{
-	    error (0, errno, "cannot chdir to %s", cp);
-	    return (1);
+	    error (0, errno, "cannot chdir to %s", dir);
+	    retval = 1;
+	    goto out;
 	}
-	if (!isfile (CVSADM) && strcmp (command_name, "export") != 0)
+	if (dirs->repository != NULL)
 	{
-	    (void) sprintf (repository, "%s/%s", prepath, path2);
-	    /* I'm not sure whether this check is redundant.  */
-	    if (!isdir (repository))
-		error (1, 0, "there is no repository %s", repository);
-	    Create_Admin (".", path, repository, sticky ? (char *) NULL : tag,
-			  sticky ? (char *) NULL : date);
-	    if (!noexec)
-	    {
-		fp = open_file (CVSADM_ENTSTAT, "w+");
-		if (fclose(fp) == EOF)
-		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
-#ifdef SERVER_SUPPORT
-		if (server_active)
-		    server_set_entstat (path, repository);
-#endif
-	    }
+	    build_one_dir (dirs->repository, dirs->dirpath, sticky);
+	    free (dirs->repository);
 	}
-	*slash = '/';
-	*slash2 = '/';
+	nextdir = dirs->next;
+	free (dirs->dirpath);
+	free (dirs);
+	dirs = nextdir;
     }
-    Subdir_Register ((List *) NULL, (char *) NULL, cp);
-    (void) CVS_MKDIR (cp, 0777);
-    if ( CVS_CHDIR (cp) < 0)
-    {
-	error (0, errno, "cannot chdir to %s", cp);
-	return (1);
-    }
-    return (0);
+
+ out:
+    return retval;
 }
