@@ -3,19 +3,14 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  */
 
 #include "cvs.h"
 
-#ifdef SERVER_SUPPORT
-static void sticky_ck PROTO((char *file, int aflag, Vers_TS * vers,
-			     List * entries,
-			     char *repository, char *update_dir));
-#else
-static void sticky_ck PROTO((char *file, int aflag, Vers_TS * vers, List * entries));
-#endif
+static void sticky_ck PROTO ((struct file_info *finfo, int aflag,
+			      Vers_TS * vers));
 
 /*
  * Classify the state of a file
@@ -26,7 +21,11 @@ Classify_File (finfo, tag, date, options, force_tag_match, aflag, versp,
     struct file_info *finfo;
     char *tag;
     char *date;
+
+    /* Keyword expansion options.  Can be either NULL or "" to
+       indicate none are specified here.  */
     char *options;
+
     int force_tag_match;
     int aflag;
     Vers_TS **versp;
@@ -48,6 +47,11 @@ Classify_File (finfo, tag, date, options, force_tag_match, aflag, versp,
 	    if (vers->ts_user == NULL)
 	    {
 		/* there is no user file */
+		/* FIXME: Why do we skip this message if vers->tag or
+		   vers->date is set?  It causes "cvs update -r tag98 foo"
+		   to silently do nothing, which is seriously confusing
+		   behavior.  "cvs update foo" gives this message, which
+		   is what I would expect.  */
 		if (!force_tag_match || !(vers->tag || vers->date))
 		    if (!really_quiet)
 			error (0, 0, "nothing known about %s", finfo->fullname);
@@ -56,10 +60,15 @@ Classify_File (finfo, tag, date, options, force_tag_match, aflag, versp,
 	    else
 	    {
 		/* there is a user file */
+		/* FIXME: Why do we skip this message if vers->tag or
+		   vers->date is set?  It causes "cvs update -r tag98 foo"
+		   to silently do nothing, which is seriously confusing
+		   behavior.  "cvs update foo" gives this message, which
+		   is what I would expect.  */
 		if (!force_tag_match || !(vers->tag || vers->date))
 		    if (!really_quiet)
-			error (0, 0, "use `cvs add' to create an entry for %s",
-			       finfo->fullname);
+			error (0, 0, "use `%s add' to create an entry for %s",
+			       program_name, finfo->fullname);
 		ret = T_UNKNOWN;
 	    }
 	}
@@ -69,8 +78,8 @@ Classify_File (finfo, tag, date, options, force_tag_match, aflag, versp,
 		ret = T_UPTODATE;
 	    else
 	    {
-		error (0, 0, "use `cvs add' to create an entry for %s",
-		       finfo->fullname);
+		error (0, 0, "use `%s add' to create an entry for %s",
+		       program_name, finfo->fullname);
 		ret = T_UNKNOWN;
 	    }
 	}
@@ -172,12 +181,8 @@ conflict: %s created independently by second party",
 
 	if (vers->ts_user == NULL)
 	{
-	    char tmp[PATH_MAX];
-
 	    /* There is no user file (as it should be) */
 
-	    (void) sprintf (tmp, "-%s", vers->vn_rcs ? vers->vn_rcs : "");
-
 	    if (vers->vn_rcs == NULL
 		|| RCS_isdead (vers->srcfile, vers->vn_rcs))
 	    {
@@ -188,8 +193,9 @@ conflict: %s created independently by second party",
 		 */
 		ret = T_REMOVE_ENTRY;
 	    }
-	    else if (strcmp (tmp, vers->vn_user) == 0)
-
+	    else if (vers->vn_rcs == NULL
+		     ? vers->vn_user[1] == '\0'
+		     : strcmp (vers->vn_rcs, vers->vn_user + 1) == 0)
 		/*
 		 * The RCS file is the same version as the user file was, and
 		 * that's OK; remove it
@@ -301,12 +307,7 @@ conflict: %s created independently by second party",
 		    ret = T_CHECKOUT;
 		else
 		{
-#ifdef SERVER_SUPPORT
-		    sticky_ck (finfo->file, aflag, vers, finfo->entries,
-			       finfo->repository, finfo->update_dir);
-#else
-		    sticky_ck (finfo->file, aflag, vers, finfo->entries);
-#endif
+		    sticky_ck (finfo, aflag, vers);
 		    ret = T_UPTODATE;
 		}
 	    }
@@ -332,12 +333,7 @@ conflict: %s created independently by second party",
 			ret = T_NEEDS_MERGE;
 #else
 		    ret = T_MODIFIED;
-#ifdef SERVER_SUPPORT
-		    sticky_ck (finfo->file, aflag, vers, finfo->entries,
-			       finfo->repository, finfo->update_dir);
-#else
-		    sticky_ck (finfo->file, aflag, vers, finfo->entries);
-#endif /* SERVER_SUPPORT */
+		    sticky_ck (finfo, aflag, vers);
 #endif
 		}
 		else
@@ -427,19 +423,10 @@ conflict: %s created independently by second party",
 }
 
 static void
-#ifdef SERVER_SUPPORT
-sticky_ck (file, aflag, vers, entries, repository, update_dir)
-#else
-sticky_ck (file, aflag, vers, entries)
-#endif
-    char *file;
+sticky_ck (finfo, aflag, vers)
+    struct file_info *finfo;
     int aflag;
     Vers_TS *vers;
-    List *entries;
-#ifdef SERVER_SUPPORT
-    char *repository;
-    char *update_dir;
-#endif
 {
     if (aflag || vers->tag || vers->date)
     {
@@ -451,7 +438,7 @@ sticky_ck (file, aflag, vers, entries)
 	    (entdate && vers->date && strcmp (entdate, vers->date)) ||
 	    ((entdate && !vers->date) || (!entdate && vers->date)))
 	{
-	    Register (entries, file, vers->vn_user, vers->ts_rcs,
+	    Register (finfo->entries, finfo->file, vers->vn_user, vers->ts_rcs,
 		      vers->options, vers->tag, vers->date, vers->ts_conflict);
 
 #ifdef SERVER_SUPPORT
@@ -461,7 +448,7 @@ sticky_ck (file, aflag, vers, entries)
 		   It is possible we will later update it again via
 		   server_updated or some such, but that is OK.  */
 		server_update_entries
-		  (file, update_dir, repository,
+		  (finfo->file, finfo->update_dir, finfo->repository,
 		   strcmp (vers->ts_rcs, vers->ts_user) == 0 ?
 		   SERVER_UPDATED : SERVER_MERGED);
 	    }
