@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Commit Files
  * 
@@ -54,7 +54,7 @@ static void fixbranch PROTO((RCSNode *, char *branch));
 static void unlockrcs PROTO((RCSNode *rcs));
 static void ci_delproc PROTO((Node *p));
 static void masterlist_delproc PROTO((Node *p));
-static void locate_rcs PROTO((char *file, char *repository, char *rcs));
+static char *locate_rcs PROTO((char *file, char *repository));
 
 struct commit_info
 {
@@ -75,6 +75,7 @@ static int run_module_prog = 1;
 static int aflag;
 static char *tag;
 static char *write_dirtag;
+static int write_dirnonbranch;
 static char *logfile;
 static List *mulist;
 static char *message;
@@ -91,6 +92,7 @@ static const char *const commit_usage[] =
     "\t-F file\tRead the log message from file.\n",
     "\t-m msg\tLog message.\n",
     "\t-r rev\tCommit to this branch or trunk revision.\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
@@ -123,6 +125,10 @@ struct find_data {
        the repository (pointer into storage managed by the recursion
        processor.  */
     char *repository;
+
+    /* Non-zero if we should force the commit.  This is enabled by
+       either -f or -r options, unlike force_ci which is just -f.  */
+    int force;
 };
 
 static Dtype find_dirent_proc PROTO ((void *callerdat, char *dir,
@@ -139,8 +145,22 @@ find_dirent_proc (callerdat, dir, repository, update_dir, entries)
 {
     struct find_data *find_data = (struct find_data *)callerdat;
 
+    /* This check seems to slowly be creeping throughout CVS (update
+       and send_dirent_proc by CVS 1.5, diff in 31 Oct 1995.  My guess
+       is that it (or some variant thereof) should go in all the
+       dirent procs.  Unless someone has some better idea...  */
+    if (!isdir (dir))
+	return (R_SKIP_ALL);
+
     /* initialize the ignore list for this directory */
     find_data->ignlist = getlist ();
+
+    /* Print the same warm fuzzy as in check_direntproc, since that
+       code will never be run during client/server operation and we
+       want the messages to match. */
+    if (!quiet)
+	error (0, 0, "Examining %s", update_dir);
+
     return R_PROCESS;
 }
 
@@ -227,7 +247,7 @@ find_fileproc (callerdat, finfo)
     xfinfo.repository = NULL;
     xfinfo.rcs = NULL;
 
-    vers = Version_TS (&xfinfo, NULL, NULL, NULL, 0, 0);
+    vers = Version_TS (&xfinfo, NULL, tag, NULL, 0, 0);
     if (vers->ts_user == NULL
 	&& vers->vn_user != NULL
 	&& vers->vn_user[0] == '-')
@@ -241,8 +261,8 @@ find_fileproc (callerdat, finfo)
 	if (vers->ts_user == NULL)
 	    error (0, 0, "nothing known about `%s'", finfo->fullname);
 	else
-	    error (0, 0, "use `cvs add' to create an entry for %s",
-		   finfo->fullname);
+	    error (0, 0, "use `%s add' to create an entry for %s",
+		   program_name, finfo->fullname);
 	return 1;
     }
     else if (vers->ts_user != NULL
@@ -254,7 +274,9 @@ find_fileproc (callerdat, finfo)
 	status = T_ADDED;
     else if (vers->ts_user != NULL
 	     && vers->ts_rcs != NULL
-	     && strcmp (vers->ts_user, vers->ts_rcs) != 0)
+	     && (args->force || strcmp (vers->ts_user, vers->ts_rcs) != 0))
+	/* If we are forcing commits, pretend that the file is
+           modified.  */
 	status = T_MODIFIED;
     else
     {
@@ -271,6 +293,7 @@ find_fileproc (callerdat, finfo)
     data = (struct logfile_info *) xmalloc (sizeof (struct logfile_info));
     data->type = status;
     data->tag = xstrdup (vers->tag);
+    data->rev_old = data->rev_new = NULL;
 
     node->type = UPDATE;
     node->delproc = update_delproc;
@@ -313,6 +336,9 @@ commit (argc, argv)
      * For log purposes, do not allow "root" to commit files.  If you look
      * like root, but are really logged in as a non-root user, it's OK.
      */
+    /* FIXME: Shouldn't this check be much more closely related to the
+       readonly user stuff (CVSROOT/readers, &c).  That is, why should
+       root be able to "cvs init", "cvs import", &c, but not "cvs ci"?  */
     if (geteuid () == (uid_t) 0)
     {
 	struct passwd *pw;
@@ -324,8 +350,8 @@ commit (argc, argv)
     }
 #endif /* CVS_BADROOT */
 
-    optind = 1;
-    while ((c = getopt (argc, argv, "nlRm:fF:r:")) != -1)
+    optind = 0;
+    while ((c = getopt (argc, argv, "+nlRm:fF:r:")) != -1)
     {
 	switch (c)
 	{
@@ -334,9 +360,9 @@ commit (argc, argv)
 		break;
 	    case 'm':
 #ifdef FORCE_USE_EDITOR
-		use_editor = TRUE;
+		use_editor = 1;
 #else
-		use_editor = FALSE;
+		use_editor = 0;
 #endif
 		if (message)
 		{
@@ -363,9 +389,9 @@ commit (argc, argv)
 		break;
 	    case 'F':
 #ifdef FORCE_USE_EDITOR
-		use_editor = TRUE;
+		use_editor = 1;
 #else
-		use_editor = FALSE;
+		use_editor = 0;
 #endif
 		logfile = optarg;
 		break;
@@ -417,6 +443,7 @@ commit (argc, argv)
 #ifdef CLIENT_SUPPORT
     if (client_active) 
     {
+	int err;
 	struct find_data find_args;
 
 	ign_setup ();
@@ -427,6 +454,12 @@ commit (argc, argv)
 	find_args.ignlist = NULL;
 	find_args.repository = NULL;
 
+	/* It is possible that only a numeric tag should set this.
+	   I haven't really thought about it much.
+	   Anyway, I suspect that setting it unnecessarily only causes
+	   a little unneeded network traffic.  */
+	find_args.force = force_ci || tag != NULL;
+
 	err = start_recursion (find_fileproc, find_filesdoneproc,
 			       find_dirent_proc, (DIRLEAVEPROC) NULL,
 			       (void *)&find_args,
@@ -469,7 +502,14 @@ commit (argc, argv)
 	if (use_editor)
 	    do_editor (".", &message, (char *)NULL, find_args.ulist);
 
+	/* Run the user-defined script to verify/check information in
+	 *the log message
+	 */
+	do_verify (message, (char *)NULL);  	
+
 	/* We always send some sort of message, even if empty.  */
+	/* FIXME: is that true?  There seems to be some code in do_editor
+	   which can leave the message NULL.  */
 	option_with_arg ("-m", message);
 
 	/* OK, now process all the questionable files we have been saving
@@ -527,10 +567,46 @@ commit (argc, argv)
 	   previous versions of client/server CVS, but it probably is a Good
 	   Thing, or at least Not Such A Bad Thing.  */
 	send_file_names (find_args.argc, find_args.argv, 0);
-	send_files (find_args.argc, find_args.argv, local, 0);
+
+	/* FIXME: This whole find_args.force/SEND_FORCE business is a
+	   kludge.  It would seem to be a server bug that we have to
+	   say that files are modified when they are not.  This makes
+	   "cvs commit -r 2" across a whole bunch of files a very slow
+	   operation (and it isn't documented in cvsclient.texi).  I
+	   haven't looked at the server code carefully enough to be
+	   _sure_ why this is needed, but if it is because the "ci"
+	   program, which we used to call, wanted the file to exist,
+	   then it would be relatively simple to fix in the server.  */
+	send_files (find_args.argc, find_args.argv, local, 0,
+		    find_args.force ? SEND_FORCE : 0);
 
 	send_to_server ("ci\012", 0);
-	return get_responses_and_close ();
+	err = get_responses_and_close ();
+	if (err != 0 && use_editor && message != NULL)
+	{
+	    /* If there was an error, don't nuke the user's carefully
+	       constructed prose.  This is something of a kludge; a better
+	       solution is probably more along the lines of #150 in TODO
+	       (doing a second up-to-date check before accepting the
+	       log message has also been suggested, but that seems kind of
+	       iffy because the real up-to-date check could still fail,
+	       another error could occur, &c.  Also, a second check would
+	       slow things down).  */
+
+	    char *fname;
+	    FILE *fp;
+
+	    fname = cvs_temp_name ();
+	    fp = CVS_FOPEN (fname, "w+");
+	    if (fp == NULL)
+		error (1, 0, "cannot create temporary file %s", fname);
+	    if (fwrite (message, 1, strlen (message), fp) != strlen (message))
+		error (1, errno, "cannot write temporary file %s", fname);
+	    if (fclose (fp) < 0)
+		error (0, errno, "cannot close temporary file %s", fname);
+	    error (0, 0, "saving log message in %s", fname);
+	}
+	return err;
     }
 #endif
 
@@ -558,13 +634,14 @@ commit (argc, argv)
 			   argv, local, W_LOCAL, aflag, 0, (char *) NULL, 1);
     if (err)
     {
-	lock_tree_cleanup ();
+	Lock_Cleanup ();
 	error (1, 0, "correct above errors first!");
     }
 
     /*
      * Run the recursion processor to commit the files
      */
+    write_dirnonbranch = 0;
     if (noexec == 0)
 	err = start_recursion (commit_fileproc, commit_filesdoneproc,
 			       commit_direntproc, commit_dirleaveproc, NULL,
@@ -574,7 +651,7 @@ commit (argc, argv)
     /*
      * Unlock all the dirs and clean up
      */
-    lock_tree_cleanup ();
+    Lock_Cleanup ();
     dellist (&mulist);
 
     if (last_register_time)
@@ -591,25 +668,21 @@ commit (argc, argv)
     return (err);
 }
 
-/*
- * Check to see if a file is ok to commit and make sure all files are
- * up-to-date
- */
-/* ARGSUSED */
-static int
-check_fileproc (callerdat, finfo)
-    void *callerdat;
-    struct file_info *finfo;
-{
-    Ctype status;
-    char *xdir;
-    Node *p;
-    List *ulist, *cilist;
-    Vers_TS *vers;
-    struct commit_info *ci;
-    struct logfile_info *li;
-    int save_noexec, save_quiet, save_really_quiet;
+/* This routine determines the status of a given file and retrieves
+   the version information that is associated with that file. */
 
+static
+Ctype
+classify_file_internal (finfo, vers)
+    struct file_info *finfo;
+    Vers_TS **vers;
+{
+    int save_noexec, save_quiet, save_really_quiet;
+    Ctype status;
+
+    /* FIXME: Do we need to save quiet as well as really_quiet?  Last
+       time I glanced at Classify_File I only saw it looking at really_quiet
+       not quiet.  */
     save_noexec = noexec;
     save_quiet = quiet;
     save_really_quiet = really_quiet;
@@ -622,15 +695,15 @@ check_fileproc (callerdat, finfo)
 	if (numdots (tag) < 2)
 	{
 	    status = Classify_File (finfo, (char *) NULL, (char *) NULL,
-				    (char *) NULL, 1, aflag, &vers, 0);
+				    (char *) NULL, 1, aflag, vers, 0);
 	    if (status == T_UPTODATE || status == T_MODIFIED ||
 		status == T_ADDED)
 	    {
 		Ctype xstatus;
 
-		freevers_ts (&vers);
+		freevers_ts (vers);
 		xstatus = Classify_File (finfo, tag, (char *) NULL,
-					 (char *) NULL, 1, aflag, &vers, 0);
+					 (char *) NULL, 1, aflag, vers, 0);
 		if (xstatus == T_REMOVE_ENTRY)
 		    status = T_MODIFIED;
 		else if (status == T_MODIFIED && xstatus == T_CONFLICT)
@@ -654,31 +727,54 @@ check_fileproc (callerdat, finfo)
 		*cp = '\0';
 	    }
 	    status = Classify_File (finfo, xtag, (char *) NULL,
-				    (char *) NULL, 1, aflag, &vers, 0);
+				    (char *) NULL, 1, aflag, vers, 0);
 	    if ((status == T_REMOVE_ENTRY || status == T_CONFLICT)
 		&& (cp = strrchr (xtag, '.')) != NULL)
 	    {
 		/* pluck one more dot off the revision */
 		*cp = '\0';
-		freevers_ts (&vers);
+		freevers_ts (vers);
 		status = Classify_File (finfo, xtag, (char *) NULL,
-					(char *) NULL, 1, aflag, &vers, 0);
+					(char *) NULL, 1, aflag, vers, 0);
 		if (status == T_UPTODATE || status == T_REMOVE_ENTRY)
 		    status = T_MODIFIED;
 	    }
 	    /* now, muck with vers to make the tag correct */
-	    free (vers->tag);
-	    vers->tag = xstrdup (tag);
+	    free ((*vers)->tag);
+	    (*vers)->tag = xstrdup (tag);
 	    free (xtag);
 	}
     }
     else
 	status = Classify_File (finfo, tag, (char *) NULL, (char *) NULL,
-				1, 0, &vers, 0);
+				1, 0, vers, 0);
     noexec = save_noexec;
     quiet = save_quiet;
     really_quiet = save_really_quiet;
 
+    return status;
+}
+
+/*
+ * Check to see if a file is ok to commit and make sure all files are
+ * up-to-date
+ */
+/* ARGSUSED */
+static int
+check_fileproc (callerdat, finfo)
+    void *callerdat;
+    struct file_info *finfo;
+{
+    Ctype status;
+    char *xdir;
+    Node *p;
+    List *ulist, *cilist;
+    Vers_TS *vers;
+    struct commit_info *ci;
+    struct logfile_info *li;
+    
+    status = classify_file_internal (finfo, &vers);
+
     /*
      * If the force-commit option is enabled, and the file in question
      * appears to be up-to-date, just make it look modified so that
@@ -768,33 +864,32 @@ check_fileproc (callerdat, finfo)
 		    return (1);
 		}
 
-		/*
-		 * If the timestamps differ, look for Conflict indicators
-		 * in the file to see if we should block the commit anyway
-		 */
-		run_setup ("%s", GREP);
-		run_arg (RCS_MERGE_PAT);
-		run_arg (finfo->file);
-		retcode = run_exec (RUN_TTY, DEVNULL, RUN_TTY, RUN_REALLY);
-		    
-		if (retcode == -1)
-		{
-		    error (1, errno,
-			   "fork failed while examining conflict in `%s'",
-			   finfo->fullname);
-		}
-		else if (retcode == 0)
+		if (file_has_markers (finfo))
 		{
+		    /* Make this a warning, not an error, because we have
+		       no way of knowing whether the "conflict indicators"
+		       are really from a conflict or whether they are part
+		       of the document itself (cvs.texinfo and sanity.sh in
+		       CVS itself, for example, tend to want to have strings
+		       like ">>>>>>>" at the start of a line).  Making people
+		       kludge this the way they need to kludge keyword
+		       expansion seems undesirable.  And it is worse than
+		       keyword expansion, because there is no -ko
+		       analogue.  */
 		    error (0, 0,
-			   "file `%s' still contains conflict indicators",
+			   "\
+warning: file `%s' seems to still contain conflict indicators",
 			   finfo->fullname);
-		    freevers_ts (&vers);
-		    return (1);
 		}
 	    }
 
 	    if (status == T_REMOVED && vers->tag && isdigit (*vers->tag))
 	    {
+		/* Remove also tries to forbid this, but we should check
+		   here.  I'm only _sure_ about somewhat obscure cases
+		   (hacking the Entries file, using an old version of
+		   CVS for the remove and a new one for the commit), but
+		   there might be other cases.  */
 		error (0, 0,
 	"cannot remove file `%s' which has a numeric sticky tag of `%s'",
 			   finfo->fullname, vers->tag);
@@ -805,7 +900,12 @@ check_fileproc (callerdat, finfo)
 	    {
 	        if (vers->tag == NULL)
 		{
-		    char rcs[PATH_MAX];
+		    char *rcs;
+
+		    rcs = xmalloc (strlen (finfo->repository)
+				   + strlen (finfo->file)
+				   + sizeof RCSEXT
+				   + 5);
 
 		    /* Don't look in the attic; if it exists there we
 		       will move it back out in checkaddfile.  */
@@ -817,8 +917,10 @@ check_fileproc (callerdat, finfo)
 		    "cannot add file `%s' when RCS file `%s' already exists",
 			       finfo->fullname, rcs);
 			freevers_ts (&vers);
+			free (rcs);
 			return (1);
 		    }
+		    free (rcs);
 		}
 		if (vers->tag && isdigit (*vers->tag) &&
 		    numdots (vers->tag) > 1)
@@ -868,6 +970,8 @@ check_fileproc (callerdat, finfo)
 		  xmalloc (sizeof (struct logfile_info)));
 	    li->type = status;
 	    li->tag = xstrdup (vers->tag);
+	    li->rev_old = xstrdup (vers->vn_rcs);
+	    li->rev_new = NULL;
 	    p->data = (char *) li;
 	    (void) addnode (ulist, p);
 
@@ -905,7 +1009,8 @@ check_fileproc (callerdat, finfo)
 }
 
 /*
- * Print warm fuzzies while examining the dirs
+ * By default, return the code that tells do_recursion to examine all
+ * directories
  */
 /* ARGSUSED */
 static Dtype
@@ -916,6 +1021,9 @@ check_direntproc (callerdat, dir, repos, update_dir, entries)
     char *update_dir;
     List *entries;
 {
+    if (!isdir (dir))
+	return (R_SKIP_ALL);
+
     if (!quiet)
 	error (0, 0, "Examining %s", update_dir);
 
@@ -972,7 +1080,8 @@ precommit_proc (repository, filter)
 	free (s);
     }
 
-    run_setup ("%s %s", filter, repository);
+    run_setup (filter);
+    run_arg (repository);
     (void) walklist (ulist, precommit_list_proc, NULL);
     return (run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL|RUN_REALLY));
 }
@@ -1017,7 +1126,7 @@ check_filesdoneproc (callerdat, err, repos, update_dir, entries)
  * Do the work of committing a file
  */
 static int maxrev;
-static char sbranch[PATH_MAX];
+static char *sbranch;
 
 /* ARGSUSED */
 static int
@@ -1030,6 +1139,21 @@ commit_fileproc (callerdat, finfo)
     List *ulist, *cilist;
     struct commit_info *ci;
 
+    /* Keep track of whether write_dirtag is a branch tag.
+       Note that if it is a branch tag in some files and a nonbranch tag
+       in others, treat it as a nonbranch tag.  It is possible that case
+       should elicit a warning or an error.  */
+    if (write_dirtag != NULL
+	&& finfo->rcs != NULL)
+    {
+	char *rev = RCS_getversion (finfo->rcs, write_dirtag, NULL, 1, NULL);
+	if (rev != NULL
+	    && !RCS_nodeisbranch (finfo->rcs, write_dirtag))
+	    write_dirnonbranch = 1;
+	if (rev != NULL)
+	    free (rev);
+    }
+
     if (finfo->update_dir[0] == '\0')
 	p = findnode (mulist, ".");
     else
@@ -1049,11 +1173,13 @@ commit_fileproc (callerdat, finfo)
      * with files as args from the command line.  In that latter case, we
      * need to get the commit message ourselves
      */
-    if (use_editor && !got_message)
-      {
+    if (!(got_message))
+    {
 	got_message = 1;
-	do_editor (finfo->update_dir, &message, finfo->repository, ulist);
-      }
+	if (use_editor)
+	    do_editor (finfo->update_dir, &message, finfo->repository, ulist);
+	do_verify (message, finfo->repository);  
+    }
 
     p = findnode (cilist, finfo->file);
     if (p == NULL)
@@ -1172,6 +1298,33 @@ out:
 	if (p)
 	    delnode (p);
     }
+    else
+    {
+	/* On success, retrieve the new version number of the file and
+           copy it into the log information (see logmsg.c
+           (logfile_write) for more details).  We should only update
+           the version number for files that have been added or
+           modified but not removed.  Why?  classify_file_internal
+           will return the version number of a file even after it has
+           been removed from the archive, which is not the behavior we
+           want for our commitlog messages; we want the old version
+           number and then "NONE." */
+	
+	if (ci->status != T_REMOVED)
+	{
+	    p = findnode (ulist, finfo->file);
+	    if (p)
+	    {
+		Vers_TS *vers;
+		struct logfile_info *li;
+	    
+		(void) classify_file_internal (finfo, &vers);
+		li = (struct logfile_info *) p->data;
+		li->rev_new = xstrdup (vers->vn_rcs);
+		freevers_ts (&vers);
+	    }
+	}
+    }
 
     return (err);
 }
@@ -1248,11 +1401,14 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
 		if (line[line_length - 1] == '\n')
 		    line[--line_length] = '\0';
 		repository = Name_Repository ((char *) NULL, update_dir);
-		run_setup ("%s %s", line, repository);
-		(void) printf ("%s %s: Executing '", program_name,
-			       command_name);
+		run_setup (line);
+		run_arg (repository);
+		cvs_output (program_name, 0);
+		cvs_output (" ", 1);
+		cvs_output (command_name, 0);
+		cvs_output (": Executing '", 0);
 		run_print (stdout);
-		(void) printf ("'\n");
+		cvs_output ("'\n", 0);
 		(void) run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL);
 		free (repository);
 	    }
@@ -1278,7 +1434,7 @@ commit_filesdoneproc (callerdat, err, repository, update_dir, entries)
 }
 
 /*
- * Get the log message for a dir and print a warm fuzzy
+ * Get the log message for a dir
  */
 /* ARGSUSED */
 static Dtype
@@ -1293,6 +1449,9 @@ commit_direntproc (callerdat, dir, repos, update_dir, entries)
     List *ulist;
     char *real_repos;
 
+    if (!isdir (dir))
+	return (R_SKIP_ALL);
+
     /* find the update list for this dir */
     p = findnode (mulist, update_dir);
     if (p != NULL)
@@ -1304,18 +1463,13 @@ commit_direntproc (callerdat, dir, repos, update_dir, entries)
     if (ulist == NULL || ulist->list->next == ulist->list)
 	return (R_SKIP_FILES);
 
-    /* print the warm fuzzy */
-    if (!quiet)
-	error (0, 0, "Committing %s", update_dir);
-
     /* get commit message */
+    real_repos = Name_Repository (dir, update_dir);
+    got_message = 1;
     if (use_editor)
-    {
-	got_message = 1;
-	real_repos = Name_Repository (dir, update_dir);
 	do_editor (update_dir, &message, real_repos, ulist);
-	free (real_repos);
-    }
+    do_verify (message, real_repos);  
+    free (real_repos);
     return (R_PROCESS);
 }
 
@@ -1332,14 +1486,13 @@ commit_dirleaveproc (callerdat, dir, err, update_dir, entries)
     List *entries;
 {
     /* update the per-directory tag info */
+    /* FIXME?  Why?  The "commit examples" node of cvs.texinfo briefly
+       mentions commit -r being sticky, but apparently in the context of
+       this being a confusing feature!  */
     if (err == 0 && write_dirtag != NULL)
     {
-	WriteTag ((char *) NULL, write_dirtag, (char *) NULL);
-#ifdef SERVER_SUPPORT
-	if (server_active)
-	    server_set_sticky (update_dir, Name_Repository (dir, update_dir),
-			       write_dirtag, (char *) NULL);
-#endif
+	WriteTag (NULL, write_dirtag, NULL, write_dirnonbranch,
+		  update_dir, Name_Repository (dir, update_dir));
     }
 
     return (err);
@@ -1404,10 +1557,10 @@ remove_file (finfo, tag, message)
 	error (1, 0, "internal error: no parsed RCS file");
 
     branch = 0;
-    if (tag && !(branch = RCS_isbranch (finfo->rcs, tag)))
+    if (tag && !(branch = RCS_nodeisbranch (finfo->rcs, tag)))
     {
 	/* a symbolic tag is specified; just remove the tag from the file */
-	if ((retcode = RCS_deltag (finfo->rcs, tag, 1)) != 0) 
+	if ((retcode = RCS_deltag (finfo->rcs, tag)) != 0) 
 	{
 	    if (!quiet)
 		error (0, retcode == -1 ? errno : 0,
@@ -1415,6 +1568,7 @@ remove_file (finfo, tag, message)
 		       finfo->fullname);
 	    return (1);
 	}
+	RCS_rewrite (finfo->rcs, NULL, NULL);
 	Scratch_Entry (finfo->entries, finfo->file);
 	return (0);
     }
@@ -1423,7 +1577,9 @@ remove_file (finfo, tag, message)
     /* commit a new, dead revision. */
 
     /* Print message indicating that file is going to be removed. */
-    (void) printf ("Removing %s;\n", finfo->fullname);
+    cvs_output ("Removing ", 0);
+    cvs_output (finfo->fullname, 0);
+    cvs_output (";\n", 0);
 
     rev = NULL;
     lockflag = 1;
@@ -1469,6 +1625,7 @@ remove_file (finfo, tag, message)
 		   finfo->fullname);
 	    return (1);
 	}
+	RCS_rewrite (finfo->rcs, NULL, NULL);
     }
 
 #ifdef SERVER_SUPPORT
@@ -1484,7 +1641,8 @@ remove_file (finfo, tag, message)
     /* check something out.  Generally this is the head.  If we have a
        particular rev, then name it.  */
     retcode = RCS_checkout (finfo->rcs, finfo->file, rev ? corev : NULL,
-			    (char *) NULL, (char *) NULL, RUN_TTY);
+			    (char *) NULL, (char *) NULL, RUN_TTY,
+			    (RCSCHECKOUTPROC) NULL, (void *) NULL);
     if (retcode != 0)
     {
 	if (!quiet)
@@ -1496,12 +1654,15 @@ remove_file (finfo, tag, message)
     /* Except when we are creating a branch, lock the revision so that
        we can check in the new revision.  */
     if (lockflag)
-	RCS_lock (finfo->rcs, rev ? corev : NULL, 0);
+    {
+	if (RCS_lock (finfo->rcs, rev ? corev : NULL, 1) == 0)
+	    RCS_rewrite (finfo->rcs, NULL, NULL);
+    }
 
     if (corev != NULL)
 	free (corev);
 
-    retcode = RCS_checkin (finfo->rcs->path, finfo->file, message, rev,
+    retcode = RCS_checkin (finfo->rcs, finfo->file, message, rev,
 			   RCS_FLAGS_DEAD | RCS_FLAGS_QUIET);
     if (retcode	!= 0)
     {
@@ -1543,10 +1704,12 @@ remove_file (finfo, tag, message)
     }
 
     /* Print message that file was removed. */
-    (void) printf ("%s  <--  %s\n", old_path, finfo->file);
-    (void) printf ("new revision: delete; ");
-    (void) printf ("previous revision: %s\n", prev_rev);
-    (void) printf ("done\n");
+    cvs_output (old_path, 0);
+    cvs_output ("  <--  ", 0);
+    cvs_output (finfo->file, 0);
+    cvs_output ("\nnew revision: delete; previous revision: ", 0);
+    cvs_output (prev_rev, 0);
+    cvs_output ("\ndone\n", 0);
     free(prev_rev);
 
     if (old_path != finfo->rcs->path)
@@ -1567,20 +1730,23 @@ finaladd (finfo, rev, tag, options)
     char *options;
 {
     int ret;
-    char tmp[PATH_MAX];
-    char rcs[PATH_MAX];
+    char *rcs;
 
-    locate_rcs (finfo->file, finfo->repository, rcs);
+    rcs = locate_rcs (finfo->file, finfo->repository);
     ret = Checkin ('A', finfo, rcs, rev, tag, options, message);
     if (ret == 0)
     {
+	char *tmp = xmalloc (strlen (finfo->file) + sizeof (CVSADM)
+			     + sizeof (CVSEXT_LOG) + 10);
 	(void) sprintf (tmp, "%s/%s%s", CVSADM, finfo->file, CVSEXT_LOG);
 	(void) unlink_file (tmp);
+	free (tmp);
     }
     else
 	fixaddfile (finfo->file, finfo->repository);
 
     (void) time (&last_register_time);
+    free (rcs);
 
     return (ret);
 }
@@ -1597,6 +1763,8 @@ unlockrcs (rcs)
     if ((retcode = RCS_unlock (rcs, NULL, 0)) != 0)
 	error (retcode == -1 ? 1 : 0, retcode == -1 ? errno : 0,
 	       "could not unlock %s", rcs->path);
+    else
+	RCS_rewrite (rcs, NULL, NULL);
 }
 
 /*
@@ -1608,10 +1776,10 @@ fixaddfile (file, repository)
     char *repository;
 {
     RCSNode *rcsfile;
-    char rcs[PATH_MAX];
+    char *rcs;
     int save_really_quiet;
 
-    locate_rcs (file, repository, rcs);
+    rcs = locate_rcs (file, repository);
     save_really_quiet = really_quiet;
     really_quiet = 1;
     if ((rcsfile = RCS_parsercsfile (rcs)) == NULL)
@@ -1619,6 +1787,7 @@ fixaddfile (file, repository)
     else
 	freercsnode (&rcsfile);
     really_quiet = save_really_quiet;
+    free (rcs);
 }
 
 /*
@@ -1631,11 +1800,12 @@ fixbranch (rcs, branch)
 {
     int retcode;
 
-    if (branch != NULL && branch[0] != '\0')
+    if (branch != NULL)
     {
 	if ((retcode = RCS_setbranch (rcs, branch)) != 0)
 	    error (retcode == -1 ? 1 : 0, retcode == -1 ? errno : 0,
 		   "cannot restore branch to %s for %s", branch, rcs->path);
+	RCS_rewrite (rcs, NULL, NULL);
     }
 }
 
@@ -1653,15 +1823,18 @@ checkaddfile (file, repository, tag, options, rcsnode)
     char *options;
     RCSNode **rcsnode;
 {
-    char rcs[PATH_MAX];
-    char fname[PATH_MAX];
+    char *rcs;
+    char *fname;
     mode_t omask;
     int retcode = 0;
     int newfile = 0;
     RCSNode *rcsfile = NULL;
+    int retval;
 
     if (tag)
     {
+	rcs = xmalloc (strlen (repository) + strlen (file)
+		       + sizeof (RCSEXT) + sizeof (CVSATTIC) + 10);
         (void) sprintf (rcs, "%s/%s%s", repository, file, RCSEXT);
 	if (! isreadable (rcs))
 	{
@@ -1675,71 +1848,132 @@ checkaddfile (file, repository, tag, options, rcsnode)
 	}
     }
     else
-	locate_rcs (file, repository, rcs);
+	rcs = locate_rcs (file, repository);
 
-    if (isreadable(rcs))
+    if (isreadable (rcs))
     {
 	/* file has existed in the past.  Prepare to resurrect. */
-	char oldfile[PATH_MAX];
 	char *rev;
 
 	if ((rcsfile = *rcsnode) == NULL)
 	{
 	    error (0, 0, "could not find parsed rcsfile %s", file);
-	    return (1);
+	    retval = 1;
+	    goto out;
 	}
 
 	if (tag == NULL)
 	{
+	    char *oldfile;
+
 	    /* we are adding on the trunk, so move the file out of the
 	       Attic. */
-	    strcpy (oldfile, rcs);
+	    oldfile = xstrdup (rcs);
 	    sprintf (rcs, "%s/%s%s", repository, file, RCSEXT);
 	    
-	    if (strcmp (oldfile, rcs) == 0
-		|| CVS_RENAME (oldfile, rcs) != 0
-		|| isreadable (oldfile)
+	    if (strcmp (oldfile, rcs) == 0)
+	    {
+		error (0, 0, "internal error: confused about attic for %s",
+		       oldfile);
+	    out1:
+		free (oldfile);
+		retval = 1;
+		goto out;
+	    }
+	    if (CVS_RENAME (oldfile, rcs) != 0)
+	    {
+		error (0, errno, "failed to move `%s' out of the attic",
+		       oldfile);
+		goto out1;
+	    }
+	    if (isreadable (oldfile)
 		|| !isreadable (rcs))
 	    {
-		error (0, 0, "failed to move `%s' out of the attic.",
-		       file);
-		return (1);
+		error (0, 0, "\
+internal error: `%s' didn't move out of the attic",
+		       oldfile);
+		goto out1;
 	    }
+	    free (oldfile);
 	    free (rcsfile->path);
 	    rcsfile->path = xstrdup (rcs);
 	}
 
 	rev = RCS_getversion (rcsfile, tag, NULL, 1, (int *) NULL);
 	/* and lock it */
-	if (lock_RCS (file, rcsfile, rev, repository)) {
+	if (lock_RCS (file, rcsfile, rev, repository))
+	{
 	    error (0, 0, "cannot lock `%s'.", rcs);
-	    free (rev);
-	    return (1);
+	    if (rev != NULL)
+		free (rev);
+	    retval = 1;
+	    goto out;
 	}
 
-	free (rev);
-    } else {
+	if (rev != NULL)
+	    free (rev);
+    }
+    else
+    {
 	/* this is the first time we have ever seen this file; create
 	   an rcs file.  */
-	run_setup ("%s%s -x,v/ -i", Rcsbin, RCS);
 
+	char *desc;
+	size_t descalloc;
+	size_t desclen;
+
+	char *opt;
+
+	desc = NULL;
+	descalloc = 0;
+	desclen = 0;
+	fname = xmalloc (strlen (file) + sizeof (CVSADM)
+			 + sizeof (CVSEXT_LOG) + 10);
 	(void) sprintf (fname, "%s/%s%s", CVSADM, file, CVSEXT_LOG);
 	/* If the file does not exist, no big deal.  In particular, the
 	   server does not (yet at least) create CVSEXT_LOG files.  */
 	if (isfile (fname))
-	    run_args ("-t%s/%s%s", CVSADM, file, CVSEXT_LOG);
+	    /* FIXME: Should be including update_dir in the appropriate
+	       place here.  */
+	    get_file (fname, fname, "r", &desc, &descalloc, &desclen);
+	free (fname);
+
+	/* From reading the RCS 5.7 source, "rcs -i" adds a newline to the
+	   end of the log message if the message is nonempty.
+	   Do it.  RCS also deletes certain whitespace, in cleanlogmsg,
+	   which we don't try to do here.  */
+	if (desclen > 0)
+	{
+	    expand_string (&desc, &descalloc, desclen + 1);
+	    desc[desclen++] = '\012';
+	}
 
 	/* Set RCS keyword expansion options.  */
 	if (options && options[0] == '-' && options[1] == 'k')
-	    run_arg (options);
-	run_arg (rcs);
-	if ((retcode = run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL)) != 0)
+	    opt = options + 2;
+	else
+	    opt = NULL;
+
+	/* This message is an artifact of the time when this
+	   was implemented via "rcs -i".  It should be revised at
+	   some point (does the "initial revision" in the message from
+	   RCS_checkin indicate that this is a new file?  Or does the
+	   "RCS file" message serve some function?).  */
+	cvs_output ("RCS file: ", 0);
+	cvs_output (rcs, 0);
+	cvs_output ("\ndone\n", 0);
+
+	if (add_rcs_file (NULL, rcs, file, NULL, opt,
+			  NULL, NULL, 0, NULL,
+			  desc, desclen, NULL) != 0)
 	{
-	    error (retcode == -1 ? 1 : 0, retcode == -1 ? errno : 0,
-		   "could not create %s", rcs);
-	    return (1);
+	    retval = 1;
+	    goto out;
 	}
+	rcsfile = RCS_parsercsfile (rcs);
 	newfile = 1;
+	if (desc != NULL)
+	    free (desc);
     }
 
     /* when adding a file for the first time, and using a tag, we need
@@ -1749,6 +1983,8 @@ checkaddfile (file, repository, tag, options, rcsnode)
 	char *tmp;
 
 	/* move the new file out of the way. */
+	fname = xmalloc (strlen (file) + sizeof (CVSADM)
+			 + sizeof (CVSPREFIX) + 10);
 	(void) sprintf (fname, "%s/%s%s", CVSADM, CVSPREFIX, file);
 	rename_file (file, fname);
 	copy_file (DEVNULL, file);
@@ -1757,25 +1993,29 @@ checkaddfile (file, repository, tag, options, rcsnode)
 	/* commit a dead revision. */
 	(void) sprintf (tmp, "file %s was initially added on branch %s.",
 			file, tag);
-	retcode = RCS_checkin (rcs, NULL, tmp, NULL,
+	retcode = RCS_checkin (rcsfile, NULL, tmp, NULL,
 			       RCS_FLAGS_DEAD | RCS_FLAGS_QUIET);
 	free (tmp);
 	if (retcode != 0)
 	{
 	    error (retcode == -1 ? 1 : 0, retcode == -1 ? errno : 0,
 		   "could not create initial dead revision %s", rcs);
-	    return (1);
+	    retval = 1;
+	    goto out;
 	}
 
 	/* put the new file back where it was */
 	rename_file (fname, file);
+	free (fname);
 
-	assert (rcsfile == NULL);
+	/* double-check that the file was written correctly */
+	freercsnode (&rcsfile);
 	rcsfile = RCS_parse (file, repository);
 	if (rcsfile == NULL)
 	{
 	    error (0, 0, "could not read %s", rcs);
-	    return (1);
+	    retval = 1;
+	    goto out;
 	}
 	if (rcsnode != NULL)
 	{
@@ -1784,9 +2024,11 @@ checkaddfile (file, repository, tag, options, rcsnode)
 	}
 
 	/* and lock it once again. */
-	if (lock_RCS (file, rcsfile, NULL, repository)) {
+	if (lock_RCS (file, rcsfile, NULL, repository))
+	{
 	    error (0, 0, "cannot lock `%s'.", rcs);
-	    return (1);
+	    retval = 1;
+	    goto out;
 	}
     }
 
@@ -1805,20 +2047,23 @@ checkaddfile (file, repository, tag, options, rcsnode)
 		if (rcsfile == NULL)
 		{
 		    error (0, 0, "could not read %s", rcs);
-		    return (1);
+		    retval = 1;
+		    goto out;
 		}
 	    }
 	}
 
-	if (!RCS_nodeisbranch (rcsfile, tag)) {
+	if (!RCS_nodeisbranch (rcsfile, tag))
+	{
 	    /* branch does not exist.  Stub it.  */
 	    char *head;
 	    char *magicrev;
-	    
+
 	    head = RCS_getversion (rcsfile, NULL, NULL, 0, (int *) NULL);
 	    magicrev = RCS_magicrev (rcsfile, head);
 
 	    retcode = RCS_settag (rcsfile, tag, magicrev);
+	    RCS_rewrite (rcsfile, NULL, NULL);
 
 	    free (head);
 	    free (magicrev);
@@ -1827,15 +2072,18 @@ checkaddfile (file, repository, tag, options, rcsnode)
 	    {
 		error (retcode == -1 ? 1 : 0, retcode == -1 ? errno : 0,
 		       "could not stub branch %s for %s", tag, rcs);
-		return (1);
+		retval = 1;
+		goto out;
 	    }
 	}
 	else
 	{
 	    /* lock the branch. (stubbed branches need not be locked.)  */
-	    if (lock_RCS (file, rcsfile, NULL, repository)) {
+	    if (lock_RCS (file, rcsfile, NULL, repository))
+	    {
 		error (0, 0, "cannot lock `%s'.", rcs);
-		return (1);
+		retval = 1;
+		goto out;
 	    }
 	} 
 
@@ -1848,8 +2096,19 @@ checkaddfile (file, repository, tag, options, rcsnode)
 
     fileattr_newfile (file);
 
+    /* I don't think fix_rcs_modes is needed any more.  In the
+       add_rcs_file case, the algorithms used by add_rcs_file and
+       fix_rcs_modes are the same, so there is no need to go through
+       it all twice.  In the other cases, I think we want to just
+       preserve the mode that the file had before we started.  That is
+       a behavior change, but I would think a desirable one.  */
     fix_rcs_modes (rcs, file);
-    return (0);
+
+    retval = 0;
+
+ out:
+    free (rcs);
+    return retval;
 }
 
 /*
@@ -1891,22 +2150,24 @@ lock_RCS (user, rcs, rev, repository)
 		return (1);
 	    }
 	}
-	err = RCS_lock(rcs, NULL, 0);
+	err = RCS_lock(rcs, NULL, 1);
     }
     else
     {
 	(void) RCS_lock(rcs, rev, 1);
     }
+    RCS_rewrite (rcs, NULL, NULL);
 
     if (err == 0)
     {
+	if (sbranch != NULL)
+	    free (sbranch);
 	if (branch)
 	{
-	    (void) strcpy (sbranch, branch);
-	    free (branch);
+	    sbranch = branch;
 	}
 	else
-	    sbranch[0] = '\0';
+	    sbranch = NULL;
 	return (0);
     }
 
@@ -1919,23 +2180,57 @@ lock_RCS (user, rcs, rev, repository)
     return (1);
 }
 
-/*
- * Called when "add"ing files to the RCS respository, as it is necessary to
- * preserve the file modes in the same fashion that RCS does.  This would be
- * automatic except that we are placing the RCS ,v file very far away from
- * the user file, and I can't seem to convince RCS of the location of the
- * user file.  So we munge it here, after the ,v file has been successfully
- * initialized with "rcs -i".
- */
+/* Called when "add"ing files to the RCS respository.  It doesn't seem to
+   be possible to get RCS to use the right mode, so we change it after
+   the fact.  */
+
 static void
 fix_rcs_modes (rcs, user)
     char *rcs;
     char *user;
 {
     struct stat sb;
+    mode_t rcs_mode;
 
-    if ( CVS_STAT (user, &sb) != -1)
-	(void) chmod (rcs, (int) sb.st_mode & ~0222);
+    if (CVS_STAT (user, &sb) < 0)
+    {
+	/* FIXME: Should be ->fullname.  */
+	error (0, errno, "warning: cannot stat %s", user);
+	return;
+    }
+
+    /* Now we compute the new mode.
+
+       The algorithm that we use is:
+
+       Write permission is always off (this is what RCS and CVS have always
+       done).
+
+       If S_IRUSR is on (user read), then the read permission of
+       the RCS file will be on.  It would seem that if this is off,
+       then other users can't do "cvs update" and such, so perhaps this
+       should be hardcoded to being on (it is a strange case, though--the
+       case in which a user file doesn't have user read permission on).
+
+       If S_IXUSR is on (user execute), then set execute permission
+       on the RCS file.  This allows other users who check out the file
+       to get the right setting for whether a shell script (for example)
+       has the executable bit set.
+
+       The result of that calculation is modified by CVSUMASK.  The
+       reason, of course, that the read and execute settings take the
+       user bit and copy it to all three bits (user, group, other), is
+       that it should be CVSUMASK, not the umask of individual users,
+       which is the sole determiner of modes in the repository.  */
+
+    rcs_mode = 0;
+    if (sb.st_mode & S_IRUSR)
+	rcs_mode |= S_IRUSR | S_IRGRP | S_IROTH;
+    if (sb.st_mode & S_IXUSR)
+	rcs_mode |= S_IXUSR | S_IXGRP | S_IXOTH;
+    rcs_mode &= ~cvsumask;
+    if (chmod (rcs, rcs_mode) < 0)
+	error (0, errno, "warning: cannot change mode of %s", rcs);
 }
 
 /*
@@ -1950,6 +2245,10 @@ update_delproc (p)
     li = (struct logfile_info *) p->data;
     if (li->tag)
 	free (li->tag);
+    if (li->rev_old)
+	free (li->rev_old);
+    if (li->rev_new)
+	free (li->rev_new);
     free (li);
 }
 
@@ -1987,15 +2286,23 @@ masterlist_delproc (p)
     free (ml);
 }
 
-/*
- * Find an RCS file in the repository.
- */
-static void
-locate_rcs (file, repository, rcs)
+/* Find an RCS file in the repository.  Most parts of CVS will want to
+   rely instead on RCS_parse which performs a similar operation and is
+   called by recurse.c which then puts the result in useful places
+   like the rcs field of struct file_info.
+
+   REPOSITORY is the repository (including the directory) and FILE is
+   the filename within that directory (without RCSEXT).  Returns a
+   newly-malloc'd array containing the absolute pathname of the RCS
+   file that was found.  */
+static char *
+locate_rcs (file, repository)
     char *file;
     char *repository;
-    char *rcs;
 {
+    char *rcs;
+
+    rcs = xmalloc (strlen (repository) + strlen (file) + sizeof (RCSEXT) + 10);
     (void) sprintf (rcs, "%s/%s%s", repository, file, RCSEXT);
     if (!isreadable (rcs))
     {
@@ -2003,4 +2310,5 @@ locate_rcs (file, repository, rcs)
 	if (!isreadable (rcs))
 	    (void) sprintf (rcs, "%s/%s%s", repository, file, RCSEXT);
     }
+    return rcs;
 }
