@@ -1,4 +1,10 @@
-/* $CVSid: @(#)cvs.h 1.86 94/10/22 $	 */
+/*
+ * Copyright (c) 1992, Brian Berliner and Jeff Polk
+ * Copyright (c) 1989-1992, Brian Berliner
+ *
+ * You may distribute under the terms of the GNU General Public License as
+ * specified in the README file that comes with the CVS kit.
+ */
 
 /*
  * basic information used in all source files
@@ -92,7 +98,10 @@ extern int errno;
 #include "rcs.h"
 
 
-/* XXX - for now this is static */
+/* This actually gets set in system.h.  Note that the _ONLY_ reason for
+   this is if various system calls (getwd, getcwd, readlink) require/want
+   us to use it.  All other parts of CVS allocate pathname buffers
+   dynamically, and we want to keep it that way.  */
 #ifndef PATH_MAX
 #ifdef MAXPATHLEN
 #define	PATH_MAX MAXPATHLEN+2
@@ -101,16 +110,8 @@ extern int errno;
 #endif
 #endif /* PATH_MAX */
 
-/*
- * Copyright (c) 1992, Brian Berliner and Jeff Polk
- * Copyright (c) 1989-1992, Brian Berliner
- *
- * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
- *
- * Definitions for the CVS Administrative directory and the files it contains.
- * Here as #define's to make changing the names a simple task.
- */
+/* Definitions for the CVS Administrative directory and the files it contains.
+   Here as #define's to make changing the names a simple task.  */
 
 #ifdef USE_VMS_FILENAMES
 #define CVSADM          "CVS"
@@ -126,6 +127,8 @@ extern int errno;
 #define CVSADM_NOTIFY   "CVS/Notify."
 #define CVSADM_NOTIFYTMP "CVS/Notify.tmp"
 #define CVSADM_BASE      "CVS/Base"
+#define CVSADM_BASEREV   "CVS/Baserev."
+#define CVSADM_BASEREVTMP "CVS/Baserev.tmp"
 #define CVSADM_TEMPLATE "CVS/Template."
 #else /* USE_VMS_FILENAMES */
 #define	CVSADM		"CVS"
@@ -143,6 +146,8 @@ extern int errno;
 /* A directory in which we store base versions of files we currently are
    editing with "cvs edit".  */
 #define CVSADM_BASE     "CVS/Base"
+#define CVSADM_BASEREV  "CVS/Baserev"
+#define CVSADM_BASEREVTMP "CVS/Baserev.tmp"
 /* File which contains the template for use in log messages.  */
 #define CVSADM_TEMPLATE "CVS/Template"
 #endif /* USE_VMS_FILENAMES */
@@ -170,6 +175,7 @@ extern int errno;
 #define CVSROOTADM_COMMITINFO	"commitinfo"
 #define CVSROOTADM_TAGINFO      "taginfo"
 #define	CVSROOTADM_EDITINFO	"editinfo"
+#define CVSROOTADM_VERIFYMSG    "verifymsg"
 #define	CVSROOTADM_HISTORY	"history"
 #define CVSROOTADM_VALTAGS	"val-tags"
 #define	CVSROOTADM_IGNORE	"cvsignore"
@@ -177,6 +183,10 @@ extern int errno;
 #define CVSROOTADM_WRAPPER	"cvswrappers"
 #define CVSROOTADM_NOTIFY	"notify"
 #define CVSROOTADM_USERS	"users"
+#define CVSROOTADM_READERS	"readers"
+#define CVSROOTADM_WRITERS	"writers"
+#define CVSROOTADM_PASSWD	"passwd"
+#define CVSROOTADM_CONFIG	"config"
 
 #define CVSNULLREPOS		"Emptydir"	/* an empty directory */
 
@@ -198,8 +208,28 @@ extern int errno;
 #define CVSDOTIGNORE	".cvsignore"
 #define CVSDOTWRAPPER   ".cvswrappers"
 
+/* Command attributes -- see function lookup_command_attribute(). */
+#define CVS_CMD_IGNORE_ADMROOT        1
+
+/* Set if CVS needs to create a CVS/Root file upon completion of this
+   command.  The name may be slightly confusing, because the flag
+   isn't really as general purpose as it seems (it is not set for cvs
+   release).  */
+
+#define CVS_CMD_USES_WORK_DIR         2
+
+#define CVS_CMD_MODIFIES_REPOSITORY   4
+
 /* miscellaneous CVS defines */
+
+/* This is the string which is at the start of the non-log-message lines
+   that we put up for the user when they edit the log message.  */
 #define	CVSEDITPREFIX	"CVS: "
+/* Number of characters in CVSEDITPREFIX to compare when deciding to strip
+   off those lines.  We don't check for the space, to accomodate users who
+   have editors which strip trailing spaces.  */
+#define CVSEDITPREFIXLEN 4
+
 #define	CVSLCKAGE	(60*60)		/* 1-hour old lock files cleaned up */
 #define	CVSLCKSLEEP	30		/* wait 30 seconds before retrying */
 #define	CVSBRANCH	"1.1.1"		/* RCS branch used for vendor srcs */
@@ -214,9 +244,6 @@ extern int errno;
 #endif
 #endif /* USE_VMS_FILENAMES */
 
-#define	FALSE		0
-#define	TRUE		1
-
 /*
  * Special tags. -rHEAD	refers to the head of an RCS file, regardless of any
  * sticky tags. -rBASE	refers to the current revision the user has checked
@@ -227,10 +254,7 @@ extern int errno;
 
 /* Environment variable used by CVS */
 #define	CVSREAD_ENV	"CVSREAD"	/* make files read-only */
-#define	CVSREAD_DFLT	FALSE		/* writable files by default */
-
-#define	RCSBIN_ENV	"RCSBIN"	/* RCS binary directory */
-/* #define	RCSBIN_DFLT		   Set by options.h */
+#define	CVSREAD_DFLT	0		/* writable files by default */
 
 #define	TMPDIR_ENV	"TMPDIR"	/* Temporary directory */
 /* #define	TMPDIR_DFLT		   Set by options.h */
@@ -257,16 +281,10 @@ extern int errno;
  */
 #define	REPOS_STRIP	"/master/"
 
-/*
- * The maximum number of files per each CVS directory. This is mainly for
- * sizing arrays statically rather than dynamically.  3000 seems plenty for
- * now.
- */
-#define	MAXFILEPERDIR	3000
-#define	MAXLINELEN	5000		/* max input line from a file */
-#define	MAXPROGLEN	30000		/* max program length to system() */
-#define	MAXLISTLEN	40000		/* For [A-Z]list holders */
-#define MAXDATELEN	50		/* max length for a date */
+/* Large enough to hold DATEFORM.  Not an arbitrary limit as long as
+   it is used for that purpose, and not to hold a string from the
+   command line, the client, etc.  */
+#define MAXDATELEN	50
 
 /* The type of an entnode.  */
 enum ent_type
@@ -280,8 +298,13 @@ struct entnode
     enum ent_type type;
     char *user;
     char *version;
+
+    /* Timestamp, or "" if none (never NULL).  */
     char *timestamp;
+
+    /* Keyword expansion options, or "" if none (never NULL).  */
     char *options;
+
     char *tag;
     char *date;
     char *conflict;
@@ -305,6 +328,8 @@ struct stickydirtag
     int aflag;
     char *tag;
     char *date;
+    int nonbranch;
+
     /* This field is set by Entries_Open() if there was subdirectory
        information; Find_Directories() uses it to see whether it needs
        to scan the directory itself.  */
@@ -331,7 +356,7 @@ typedef enum direnter_type Dtype;
 #endif
 
 extern char *program_name, *program_path, *command_name;
-extern char *Rcsbin, *Tmpdir, *Editor;
+extern char *Tmpdir, *Editor;
 extern int cvsadmin_root;
 extern char *CurDir;
 extern int really_quiet, quiet;
@@ -341,7 +366,8 @@ extern mode_t cvsumask;
 
 /* Access method specified in CVSroot. */
 typedef enum {
-  local_method, server_method, pserver_method, kserver_method, ext_method
+  local_method, server_method, pserver_method, kserver_method, gserver_method,
+  ext_method
 } CVSmethod;
 extern char *method_names[];	/* change this in root.c if you change
 				   the enum above */
@@ -353,30 +379,40 @@ extern char *CVSroot_username;	/* the username or NULL if method == local */
 extern char *CVSroot_hostname;	/* the hostname or NULL if method == local */
 extern char *CVSroot_directory;	/* the directory name */
 
+extern char *emptydir_name PROTO ((void));
+
 extern int trace;		/* Show all commands */
 extern int noexec;		/* Don't modify disk anywhere */
 extern int logoff;		/* Don't write history entry */
 
+#ifdef AUTH_SERVER_SUPPORT
+extern char *Pserver_Repos;     /* used to check that same repos is
+                                   transmitted in pserver auth and in
+                                   CVS protocol. */
+#endif /* AUTH_SERVER_SUPPORT */
+
 extern char hostname[];
 
 /* Externs that are included directly in the CVS sources */
 
-int RCS_exec_settag PROTO((const char *, const char *, const char *));
-int RCS_exec_deltag PROTO((const char *, const char *, int));
-int RCS_exec_setbranch PROTO((const char *, const char *));
-int RCS_exec_lock PROTO((const char *, const char *, int));
-int RCS_exec_unlock PROTO((const char *, const char *, int));
-int RCS_merge PROTO((const char *, const char *, const char *, const char *));
-int RCS_exec_checkout PROTO ((char *rcsfile, char *workfile, char *tag,
-			      char *options, char *sout));
+int RCS_merge PROTO((RCSNode *, char *, char *, char *, char *, char *));
 /* Flags used by RCS_* functions.  See the description of the individual
    functions for which flags mean what for each function.  */
 #define RCS_FLAGS_FORCE 1
 #define RCS_FLAGS_DEAD 2
 #define RCS_FLAGS_QUIET 4
 #define RCS_FLAGS_MODTIME 8
-int RCS_checkin PROTO ((char *rcsfile, char *workfile, char *message,
-			char *rev, int flags));
+
+extern int RCS_exec_rcsdiff PROTO ((RCSNode *rcsfile,
+				    char *opts, char *options,
+				    char *rev1, char *rev2,
+				    char *label1, char *label2,
+				    char *workfile));
+extern int diff_exec PROTO ((char *file1, char *file2, char *options,
+			     char *out));
+extern int diff_execv PROTO ((char *file1, char *file2,
+			      char *label1, char *label2,
+			      char *options, char *out));
 
 
 
@@ -390,27 +426,38 @@ List *Entries_Open PROTO((int aflag));
 void Subdirs_Known PROTO((List *entries));
 void Subdir_Register PROTO((List *, const char *, const char *));
 void Subdir_Deregister PROTO((List *, const char *, const char *));
+
 char *Make_Date PROTO((char *rawdate));
 char *Name_Repository PROTO((char *dir, char *update_dir));
+char *Short_Repository PROTO((char *repository));
+void Sanitize_Repository_Name PROTO((char *repository));
+
 char *Name_Root PROTO((char *dir, char *update_dir));
 int parse_cvsroot PROTO((char *CVSroot));
 void set_local_cvsroot PROTO((char *dir));
 void Create_Root PROTO((char *dir, char *rootdir));
-int same_directories PROTO((char *dir1, char *dir2));
-char *Short_Repository PROTO((char *repository));
-char *gca PROTO((char *rev1, char *rev2));
+void root_allow_add PROTO ((char *));
+void root_allow_free PROTO ((void));
+int root_allow_ok PROTO ((char *));
+
+char *gca PROTO((const char *rev1, const char *rev2));
+extern void check_numeric PROTO ((const char *, int, char **));
 char *getcaller PROTO((void));
 char *time_stamp PROTO((char *file));
+
 char *xmalloc PROTO((size_t bytes));
 void *xrealloc PROTO((void *ptr, size_t bytes));
+void expand_string PROTO ((char **, size_t *, size_t));
 char *xstrdup PROTO((const char *str));
 void strip_trailing_newlines PROTO((char *str));
+int pathname_levels PROTO ((char *path));
+
 typedef	int (*CALLPROC)	PROTO((char *repository, char *value));
 int Parse_Info PROTO((char *infofile, char *repository, CALLPROC callproc, int all));
-int Reader_Lock PROTO((char *xrepository));
+extern int parse_config PROTO ((char *));
+
 typedef	RETSIGTYPE (*SIGCLEANUPPROC)	PROTO(());
 int SIG_register PROTO((int sig, SIGCLEANUPPROC sigcleanup));
-int Writer_Lock PROTO((List * list));
 int isdir PROTO((const char *file));
 int isfile PROTO((const char *file));
 int islink PROTO((const char *file));
@@ -423,6 +470,8 @@ char *get_homedir PROTO ((void));
 char *cvs_temp_name PROTO ((void));
 
 int numdots PROTO((const char *s));
+char *increment_revnum PROTO ((const char *));
+int compare_revnums PROTO ((const char *, const char *));
 int unlink_file PROTO((const char *f));
 int link_file PROTO ((const char *from, const char *to));
 int unlink_file_dir PROTO((const char *f));
@@ -431,27 +480,31 @@ int xcmp PROTO((const char *file1, const char *file2));
 int yesno PROTO((void));
 void *valloc PROTO((size_t bytes));
 time_t get_date PROTO((char *date, struct timeb *now));
-void Create_Admin PROTO((char *dir, char *update_dir,
-			 char *repository, char *tag, char *date));
+extern int Create_Admin PROTO ((char *dir, char *update_dir,
+				char *repository, char *tag, char *date,
+				int nonbranch, int warn));
+extern int expand_at_signs PROTO ((char *, off_t, FILE *));
 
+/* Locking subsystem (implemented in lock.c).  */
+
+int Reader_Lock PROTO((char *xrepository));
 void Lock_Cleanup PROTO((void));
 
 /* Writelock an entire subtree, well the part specified by ARGC, ARGV, LOCAL,
    and AFLAG, anyway.  */
 void lock_tree_for_write PROTO ((int argc, char **argv, int local, int aflag));
 
-/* Remove locks set by lock_tree_for_write.  Currently removes readlocks
-   too.  */
-void lock_tree_cleanup PROTO ((void));
+/* See lock.c for description.  */
+extern void lock_dir_for_write PROTO ((char *));
 
-void ParseTag PROTO((char **tagp, char **datep));
 void Scratch_Entry PROTO((List * list, char *fname));
-void WriteTag PROTO((char *dir, char *tag, char *date));
+void ParseTag PROTO((char **tagp, char **datep, int *nonbranchp));
+void WriteTag PROTO ((char *dir, char *tag, char *date, int nonbranch,
+		      char *update_dir, char *repository));
 void cat_module PROTO((int status));
 void check_entries PROTO((char *dir));
 void close_module PROTO((DBM * db));
 void copy_file PROTO((const char *from, const char *to));
-void (*error_set_cleanup PROTO((void (*) (void)))) PROTO ((void));
 void fperror PROTO((FILE * fp, int status, int errnum, char *message,...));
 void free_names PROTO((int *pargc, char *argv[]));
 
@@ -468,7 +521,7 @@ extern int ign_case;
 
 #include "update.h"
 
-void line2argv PROTO((int *pargc, char *argv[], char *line));
+void line2argv PROTO ((int *pargc, char ***argv, char *line, char *sepchars));
 void make_directories PROTO((const char *name));
 void make_directory PROTO((const char *name));
 extern int mkdir_if_needed PROTO ((char *name));
@@ -483,7 +536,11 @@ void rename_file PROTO((const char *from, const char *to));
 extern void expand_wild PROTO ((int argc, char **argv, 
                                 int *pargc, char ***pargv));
 
-void strip_path PROTO((char *path));
+#ifdef SERVER_SUPPORT
+extern int cvs_casecmp PROTO ((char *, char *));
+extern int fopen_case PROTO ((char *, char *, FILE **, char **));
+#endif
+
 void strip_trailing_slashes PROTO((char *path));
 void update_delproc PROTO((Node * p));
 void usage PROTO((const char *const *cpp));
@@ -498,8 +555,10 @@ void Update_Logfile PROTO((char *repository, char *xmessage, FILE * xlogfp,
 void do_editor PROTO((char *dir, char **messagep,
 		      char *repository, List * changes));
 
+void do_verify PROTO((char *message, char *repository));
+
 typedef	int (*CALLBACKPROC)	PROTO((int *pargc, char *argv[], char *where,
-	char *mwhere, char *mfile, int horten, int local_specified,
+	char *mwhere, char *mfile, int shorten, int local_specified,
 	char *omodule, char *msg));
 
 /* This is the structure that the recursion processor passes to the
@@ -560,6 +619,9 @@ void SIG_endCrSect PROTO((void));
 void read_cvsrc PROTO((int *argc, char ***argv, char *cmdname));
 
 char *make_message_rcslegal PROTO((char *message));
+extern int file_has_markers PROTO ((const struct file_info *));
+extern void get_file PROTO ((const char *, const char *, const char *,
+			     char **, size_t *, size_t *));
 
 /* flags for run_exec(), the fast system() for CVS */
 #define	RUN_NORMAL		0x0000	/* no special behaviour */
@@ -572,14 +634,9 @@ char *make_message_rcslegal PROTO((char *message));
 
 void run_arg PROTO((const char *s));
 void run_print PROTO((FILE * fp));
-#ifdef HAVE_VPRINTF
-void run_setup PROTO((const char *fmt,...));
-void run_args PROTO((const char *fmt,...));
-#else
-void run_setup ();
-void run_args ();
-#endif
-int run_exec PROTO((char *stin, char *stout, char *sterr, int flags));
+void run_setup PROTO ((const char *prog));
+int run_exec PROTO((const char *stin, const char *stout, const char *sterr,
+		    int flags));
 
 /* other similar-minded stuff from run.c.  */
 FILE *run_popen PROTO((const char *, const char *));
@@ -599,10 +656,16 @@ pid_t waitpid PROTO((pid_t, int *, int));
 struct vers_ts
 {
     /* rcs version user file derives from, from CVS/Entries.
-     * it can have the following special values:
-     *    empty = no user file
-     *    0 = user file is new
-     *    -vers = user file to be removed.  */
+       It can have the following special values:
+
+       NULL = file is not mentioned in Entries (this is also used for a
+	      directory).
+       "" = ILLEGAL!  The comment used to say that it meant "no user file"
+	    but as far as I know CVS didn't actually use it that way.
+	    Note that according to cvs.texinfo, "" is not legal in the
+	    Entries file.
+       0 = user file is new
+       -vers = user file to be removed.  */
     char *vn_user;
 
     /* Numeric revision number corresponding to ->vn_tag (->vn_tag
@@ -615,7 +678,9 @@ struct vers_ts
     char *vn_tag;
 
     /* This is the timestamp from stating the file in the working directory.
-       It is NULL if there is no file in the working directory.  */
+       It is NULL if there is no file in the working directory.  It is
+       "Is-modified" if we know the file is modified but don't have its
+       contents.  */
     char *ts_user;
     /* Timestamp from CVS/Entries.  For the server, ts_user and ts_rcs
        are computed in a slightly different way, but the fact remains that
@@ -623,7 +688,8 @@ struct vers_ts
        and if they differ it is modified.  */
     char *ts_rcs;
 
-    /* Options from CVS/Entries (keyword expansion).  */
+    /* Options from CVS/Entries (keyword expansion), malloc'd.  If none,
+       then it is an empty string (never NULL).  */
     char *options;
 
     /* If non-NULL, there was a conflict (or merely a merge?  See merge_file)
@@ -638,6 +704,9 @@ struct vers_ts
     /* Date specified on the command line, or if none, date stored in
        CVS/Entries.  */
     char *date;
+    /* If this is 1, then tag is not a branch tag.  If this is 0, then
+       tag may or may not be a branch tag.  */
+    int nonbranch;
 
     /* Pointer to entries file node  */
     Entnode *entdata;
@@ -659,6 +728,11 @@ int Checkin PROTO ((int type, struct file_info *finfo, char *rcs, char *rev,
 		    char *tag, char *options, char *message));
 int No_Difference PROTO ((struct file_info *finfo, Vers_TS *vers));
 
+/* CVSADM_BASEREV stuff, from entries.c.  */
+extern char *base_get PROTO ((struct file_info *));
+extern void base_register PROTO ((struct file_info *, char *));
+extern void base_deregister PROTO ((struct file_info *));
+
 /*
  * defines for Classify_File() to determine the current state of a file.
  * These are also used as types in the data field for the list we make for
@@ -694,6 +768,10 @@ struct logfile_info
 {
   enum classify_type type;
   char *tag;
+  char *rev_old;		/* rev number before a commit/modify,
+				   NULL for add or import */
+  char *rev_new;		/* rev number after a commit/modify,
+				   add, or import, NULL for remove */
 };
 
 /* Wrappers.  */
@@ -712,7 +790,7 @@ int   wrap_name_has PROTO((const char *name,WrapMergeHas has));
 char *wrap_rcsoption PROTO ((const char *fileName, int asFlag));
 char *wrap_tocvs_process_file PROTO((const char *fileName));
 int   wrap_merge_is_copy PROTO((const char *fileName));
-char *wrap_fromcvs_process_file PROTO((const char *fileName));
+void wrap_fromcvs_process_file PROTO ((const char *fileName));
 void wrap_add_file PROTO((const char *file,int temp));
 void wrap_add PROTO((char *line,int temp));
 void wrap_send PROTO ((void));
@@ -731,6 +809,26 @@ int unedit PROTO ((int argc, char **argv));
 int editors PROTO ((int argc, char **argv));
 int watchers PROTO ((int argc, char **argv));
 extern int annotate PROTO ((int argc, char **argv));
+extern int add PROTO ((int argc, char **argv));
+extern int admin PROTO ((int argc, char **argv));
+extern int checkout PROTO ((int argc, char **argv));
+extern int commit PROTO ((int argc, char **argv));
+extern int diff PROTO ((int argc, char **argv));
+extern int history PROTO ((int argc, char **argv));
+extern int import PROTO ((int argc, char **argv));
+extern int cvslog PROTO ((int argc, char **argv));
+#ifdef AUTH_CLIENT_SUPPORT
+extern int login PROTO((int argc, char **argv));
+int logout PROTO((int argc, char **argv));
+#endif /* AUTH_CLIENT_SUPPORT */
+extern int patch PROTO((int argc, char **argv));
+extern int release PROTO((int argc, char **argv));
+extern int cvsremove PROTO((int argc, char **argv));
+extern int rtag PROTO((int argc, char **argv));
+extern int status PROTO((int argc, char **argv));
+extern int cvstag PROTO((int argc, char **argv));
+
+extern unsigned long int lookup_command_attribute PROTO((char *));
 
 #if defined(AUTH_CLIENT_SUPPORT) || defined(AUTH_SERVER_SUPPORT)
 char *scramble PROTO ((char *str));
@@ -744,12 +842,14 @@ char *get_cvs_password PROTO((void));
 extern void tag_check_valid PROTO ((char *, int, char **, int, int, char *));
 extern void tag_check_valid_join PROTO ((char *, int, char **, int, int,
 					 char *));
-extern void tag_lockdir PROTO ((char *));
-extern void tag_unlockdir PROTO ((void));
 
+/* From server.c and documented there.  */
 extern void cvs_output PROTO ((const char *, size_t));
+extern void cvs_output_binary PROTO ((char *, size_t));
 extern void cvs_outerr PROTO ((const char *, size_t));
 extern void cvs_flusherr PROTO ((void));
+extern void cvs_flushout PROTO ((void));
+extern void cvs_output_tagged PROTO ((char *, char *));
 
 #if defined(SERVER_SUPPORT) || defined(CLIENT_SUPPORT)
 #include "server.h"
