@@ -1,10 +1,10 @@
 /*
- *    Copyright (c) 1993 david d zuhn
- *
- *    written by david d `zoo' zuhn while at Cygnus Support
- *
- *    You may distribute under the terms of the GNU General Public License
- *    as specified in the README file that comes with the CVS 1.4 kit.
+ * Copyright (c) 1993 david d zuhn
+ * 
+ * Written by david d `zoo' zuhn while at Cygnus Support
+ * 
+ * You may distribute under the terms of the GNU General Public License as
+ * specified in the README file that comes with the CVS source distribution.
  *
  */
 
@@ -51,17 +51,17 @@ read_cvsrc (argc, argv, cmdname)
     int max_new_argv;
     char **new_argv;
 
+    /* old_argc and old_argv hold the values returned from the
+       previous invocation of read_cvsrc and are used to free the
+       allocated memory.  The first invocation of read_cvsrc gets argv
+       from the system, this memory must not be free'd.  */
+    static int old_argc = 0;
+    static char **old_argv = NULL;
+
     /* don't do anything if argc is -1, since that implies "help" mode */
     if (*argc == -1)
 	return;
 
-    /* setup the new options list */
-
-    new_argc = 1;
-    max_new_argv = (*argc) + GROW;
-    new_argv = (char **) xmalloc (max_new_argv * sizeof (char*));
-    new_argv[0] = xstrdup ((*argv)[0]);
-
     /* determine filename for ~/.cvsrc */
 
     homedir = get_homedir ();
@@ -105,28 +105,27 @@ read_cvsrc (argc, argv, cmdname)
 
     fclose (cvsrcfile);
 
+    /* setup the new options list */
+
+    new_argc = 1;
+    max_new_argv = (*argc) + GROW;
+    new_argv = (char **) xmalloc (max_new_argv * sizeof (char*));
+    new_argv[0] = xstrdup ((*argv)[0]);
+
     if (found)
     {
 	/* skip over command in the options line */
-	optstart = strtok (line + command_len, "\t \n");
-
-	while (optstart)
+	for (optstart = strtok (line + command_len, "\t \n");
+	     optstart;
+	     optstart = strtok (NULL, "\t \n"))
 	{
-	    new_argv [new_argc] = xstrdup (optstart);
-	    new_argv [new_argc+1] = NULL;
-	    new_argc += 1;
+	    new_argv [new_argc++] = xstrdup (optstart);
 	  
 	    if (new_argc >= max_new_argv)
 	    {
-		char **tmp_argv;
 		max_new_argv += GROW;
-		tmp_argv = (char **) xmalloc (max_new_argv * sizeof (char*));
-		for (i = 0; i <= new_argc; i++)
-		    tmp_argv[i] = new_argv[i];
-		free(new_argv);
-		new_argv = tmp_argv;
+		new_argv = (char **) xrealloc (new_argv, max_new_argv * sizeof (char*));
 	    }
-	    optstart = strtok (NULL, "\t \n");
 	}
     }
 
@@ -135,14 +134,25 @@ read_cvsrc (argc, argv, cmdname)
 
     /* now copy the remaining arguments */
   
+    if (new_argc + *argc > max_new_argv)
+    {
+	max_new_argv = new_argc + *argc;
+	new_argv = (char **) xrealloc (new_argv, max_new_argv * sizeof (char*));
+    }
     for (i=1; i < *argc; i++)
     {
-	new_argv [new_argc] = (*argv)[i];
-	new_argc += 1;
+	new_argv [new_argc++] = xstrdup ((*argv)[i]);
     }
 
-    *argc = new_argc;
-    *argv = new_argv;
+    if (old_argv != NULL)
+    {
+	/* Free the memory which was allocated in the previous
+           read_cvsrc call.  */
+	free_names (&old_argc, old_argv);
+    }
+
+    old_argc = *argc = new_argc;
+    old_argv = *argv = new_argv;
 
     free (homeinit);
     return;
