@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Find Names
  * 
@@ -71,7 +71,6 @@ Find_Names (repository, which, aflag, optentries)
 {
     List *entries;
     List *files;
-    char dir[PATH_MAX];
 
     /* make a list for the files */
     files = filelist = getlist ();
@@ -103,8 +102,11 @@ Find_Names (repository, which, aflag, optentries)
 	/* search the attic too */
 	if (which & W_ATTIC)
 	{
+	    char *dir;
+	    dir = xmalloc (strlen (repository) + sizeof (CVSATTIC) + 10);
 	    (void) sprintf (dir, "%s/%s", repository, CVSATTIC);
 	    (void) find_rcs (dir, files);
+	    free (dir);
 	}
     }
 
@@ -198,6 +200,15 @@ Find_Directories (repository, which, entries)
                information is not recorded in the Entries file.  Find
                the subdirectories the hard way, and, if possible, add
                it to the Entries file for next time.  */
+
+	    /* FIXME-maybe: find_dirs is bogus for this usage because
+	       it skips CVSATTIC and CVSLCK directories--those names
+	       should be special only in the repository.  However, in
+	       the interests of not perturbing this code, we probably
+	       should leave well enough alone unless we want to write
+	       a sanity.sh test case (which would operate by manually
+	       hacking on the CVS/Entries file).  */
+
 	    if (find_dirs (".", dirlist, 1, tmpentries) != 0)
 		error (1, errno, "cannot open current directory");
 	    if (tmpentries != NULL)
@@ -221,16 +232,11 @@ Find_Directories (repository, which, entries)
 	if (find_dirs (repository, dirlist, 0, entries) != 0)
 	    error (1, errno, "cannot open directory %s", repository);
 
-#ifdef ATTIC_DIR_SUPPORT		/* XXX - FIXME */
-	/* search the attic too */
-	if (which & W_ATTIC)
-	{
-	    char dir[PATH_MAX];
-
-	    (void) sprintf (dir, "%s/%s", repository, CVSATTIC);
-	    (void) find_dirs (dir, dirlist, 0, entries);
-	}
-#endif
+	/* We don't need to look in the attic because directories
+	   never go in the attic.  In the future, there hopefully will
+	   be a better mechanism for detecting whether a directory in
+	   the repository is alive or dead; it may or may not involve
+	   moving directories to the attic.  */
     }
 
     /* sort the list into alphabetical order and return it */
@@ -259,7 +265,7 @@ find_rcs (dir, list)
     /* read the dir, grabbing the ,v files */
     while ((dp = readdir (dirp)) != NULL)
     {
-	if (fnmatch (RCSPAT, dp->d_name, 0) == 0) 
+	if (CVS_FNMATCH (RCSPAT, dp->d_name, 0) == 0) 
 	{
 	    char *comma;
 
@@ -291,9 +297,20 @@ find_dirs (dir, list, checkadm, entries)
     List *entries;
 {
     Node *p;
-    char tmp[PATH_MAX];
+    char *tmp = NULL;
+    size_t tmp_size = 0;
     struct dirent *dp;
     DIR *dirp;
+    int skip_emptydir = 0;
+
+    /* First figure out whether we need to skip directories named
+       Emptydir.  Except in the CVSNULLREPOS case, Emptydir is just
+       a normal directory name.  */
+    if (isabsolute (dir)
+	&& strncmp (dir, CVSroot_directory, strlen (CVSroot_directory)) == 0
+	&& ISDIRSEP (dir[strlen (CVSroot_directory)])
+	&& strcmp (dir + strlen (CVSroot_directory) + 1, CVSROOTADM) == 0)
+	skip_emptydir = 1;
 
     /* set up to read the dir */
     if ((dirp = CVS_OPENDIR (dir)) == NULL)
@@ -315,6 +332,10 @@ find_dirs (dir, list, checkadm, entries)
 	if (entries != NULL && findnode (entries, dp->d_name) != NULL)
 	    continue;
 
+	if (skip_emptydir
+	    && strcmp (dp->d_name, CVSNULLREPOS) == 0)
+	    continue;
+
 #ifdef DT_DIR
 	if (dp->d_type != DT_DIR) 
 	{
@@ -322,9 +343,12 @@ find_dirs (dir, list, checkadm, entries)
 		continue;
 #endif
 	    /* don't bother stating ,v files */
-	    if (fnmatch (RCSPAT, dp->d_name, 0) == 0)
+	    if (CVS_FNMATCH (RCSPAT, dp->d_name, 0) == 0)
 		continue;
 
+	    expand_string (&tmp,
+			   &tmp_size,
+			   strlen (dir) + strlen (dp->d_name) + 10);
 	    sprintf (tmp, "%s/%s", dir, dp->d_name);
 	    if (!isdir (tmp))
 		continue;
@@ -344,6 +368,8 @@ find_dirs (dir, list, checkadm, entries)
 		if (dp->d_type == DT_LNK)
 		    continue;
 #endif
+		/* Note that we only get here if we already set tmp
+		   above.  */
 		if (islink (tmp))
 		    continue;
 #ifdef DT_DIR
@@ -351,6 +377,10 @@ find_dirs (dir, list, checkadm, entries)
 #endif
 
 	    /* check for new style */
+	    expand_string (&tmp,
+			   &tmp_size,
+			   (strlen (dir) + strlen (dp->d_name)
+			    + sizeof (CVSADM) + 10));
 	    (void) sprintf (tmp, "%s/%s/%s", dir, dp->d_name, CVSADM);
 	    if (!isdir (tmp))
 		continue;
@@ -364,5 +394,7 @@ find_dirs (dir, list, checkadm, entries)
 	    freenode (p);
     }
     (void) closedir (dirp);
+    if (tmp != NULL)
+	free (tmp);
     return (0);
 }
