@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Patch
  * 
@@ -42,10 +42,11 @@ static int unidiff = 0;
 
 static const char *const patch_usage[] =
 {
-    "Usage: %s %s [-fl] [-c|-u] [-s|-t] [-V %%d]\n",
+    "Usage: %s %s [-flR] [-c|-u] [-s|-t] [-V %%d]\n",
     "    -r rev|-D date [-r rev2 | -D date2] modules...\n",
     "\t-f\tForce a head revision match if tag/date not found.\n",
     "\t-l\tLocal directory only, not recursive\n",
+    "\t-R\tProcess directories recursively.\n",
     "\t-c\tContext diffs (default)\n",
     "\t-u\tUnidiff format.\n",
     "\t-s\tShort patch - one liner per file.\n",
@@ -53,6 +54,7 @@ static const char *const patch_usage[] =
     "\t-D date\tDate.\n",
     "\t-r rev\tRevision - symbolic or numeric.\n",
     "\t-V vers\tUse RCS Version \"vers\" for keyword expansion.\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
@@ -69,8 +71,8 @@ patch (argc, argv)
     if (argc == -1)
 	usage (patch_usage);
 
-    optind = 1;
-    while ((c = getopt (argc, argv, "V:k:cuftsQqlRD:r:")) != -1)
+    optind = 0;
+    while ((c = getopt (argc, argv, "+V:k:cuftsQqlRD:r:")) != -1)
     {
 	switch (c)
 	{
@@ -124,12 +126,31 @@ patch (argc, argv)
 		options = RCS_check_kflag (optarg);
 		break;
 	    case 'V':
+		/* This option is pretty seriously broken:
+		   1.  It is not clear what it does (does it change keyword
+		   expansion behavior?  If so, how?  Or does it have
+		   something to do with what version of RCS we are using?
+		   Or the format we write RCS files in?).
+		   2.  Because both it and -k use the options variable,
+		   specifying both -V and -k doesn't work.
+		   3.  At least as of CVS 1.9, it doesn't work (failed
+		   assertion in RCS_checkout where it asserts that options
+		   starts with -k).  Few people seem to be complaining.
+		   In the future (perhaps the near future), I have in mind
+		   removing it entirely, and updating NEWS and cvs.texinfo,
+		   but in case it is a good idea to give people more time
+		   to complain if they would miss it, I'll just add this
+		   quick and dirty error message for now.  */
+		error (1, 0,
+		       "the -V option is obsolete and should not be used");
+#if 0
 		if (atoi (optarg) <= 0)
 		    error (1, 0, "must specify a version number to -V");
 		if (options)
 		    free (options);
 		options = xmalloc (strlen (optarg) + 1 + 2);	/* for the -V */
 		(void) sprintf (options, "-V%s", optarg);
+#endif
 		break;
 	    case 'u':
 		unidiff = 1;		/* Unidiff */
@@ -177,7 +198,7 @@ patch (argc, argv)
 
 	if (local)
 	    send_arg("-l");
-	if (force_tag_match)
+	if (!force_tag_match)
 	    send_arg("-f");
 	if (toptwo_diffs)
 	    send_arg("-t");
@@ -239,7 +260,6 @@ patch (argc, argv)
  * callback proc for doing the real work of patching
  */
 /* ARGSUSED */
-static char where[PATH_MAX];
 static int
 patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 	    mname, msg)
@@ -255,16 +275,21 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 {
     int err = 0;
     int which;
-    char repository[PATH_MAX];
+    char *repository;
+    char *where;
 
+    repository = xmalloc (strlen (CVSroot_directory) + strlen (argv[0])
+			  + (mfile == NULL ? 0 : strlen (mfile)) + 30);
     (void) sprintf (repository, "%s/%s", CVSroot_directory, argv[0]);
+    where = xmalloc (strlen (argv[0]) + (mfile == NULL ? 0 : strlen (mfile))
+		     + 10);
     (void) strcpy (where, argv[0]);
 
     /* if mfile isn't null, we need to set up to do only part of the module */
     if (mfile != NULL)
     {
 	char *cp;
-	char path[PATH_MAX];
+	char *path;
 
 	/* if the portion of the module is a path, put the dir part on repos */
 	if ((cp = strrchr (mfile, '/')) != NULL)
@@ -278,6 +303,7 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 	}
 
 	/* take care of the rest */
+	path = xmalloc (strlen (repository) + strlen (mfile) + 5);
 	(void) sprintf (path, "%s/%s", repository, mfile);
 	if (isdir (path))
 	{
@@ -296,14 +322,17 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 	    argv[1] = xstrdup (mfile);
 	    (*pargc) = 2;
 	}
+	free (path);
     }
 
     /* cd to the starting repository */
     if ( CVS_CHDIR (repository) < 0)
     {
 	error (0, errno, "cannot chdir to %s", repository);
+	free (repository);
 	return (1);
     }
+    free (repository);
 
     if (force_tag_match)
 	which = W_REPOS | W_ATTIC;
@@ -326,6 +355,7 @@ patch_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 			   (DIRLEAVEPROC) NULL, NULL,
 			   *pargc - 1, argv + 1, local,
 			   which, 0, 1, where, 1);
+    free (where);
 
     return (err);
 }
@@ -342,26 +372,36 @@ patch_fileproc (callerdat, finfo)
 {
     struct utimbuf t;
     char *vers_tag, *vers_head;
-    char rcsspace[1][PATH_MAX];
-    char *rcs = rcsspace[0];
+    char *rcs = NULL;
     RCSNode *rcsfile;
     FILE *fp1, *fp2, *fp3;
     int ret = 0;
     int isattic = 0;
     int retcode = 0;
-    char file1[PATH_MAX], file2[PATH_MAX], strippath[PATH_MAX];
+    char *file1;
+    char *file2;
+    char *strippath;
     char *line1, *line2;
     size_t line1_chars_allocated;
     size_t line2_chars_allocated;
     char *cp1, *cp2;
     FILE *fp;
 
+    line1 = NULL;
+    line1_chars_allocated = 0;
+    line2 = NULL;
+    line2_chars_allocated = 0;
+
     /* find the parsed rcs file */
     if ((rcsfile = finfo->rcs) == NULL)
-	return (1);
+    {
+	ret = 1;
+	goto out2;
+    }
     if ((rcsfile->flags & VALID) && (rcsfile->flags & INATTIC))
 	isattic = 1;
 
+    rcs = xmalloc (strlen (finfo->file) + sizeof (RCSEXT) + 5);
     (void) sprintf (rcs, "%s%s", finfo->file, RCSEXT);
 
     /* if vers_head is NULL, may have been removed from the release */
@@ -381,17 +421,21 @@ patch_fileproc (callerdat, finfo)
     if (toptwo_diffs)
     {
 	if (vers_head == NULL)
-	    return (1);
+	{
+	    ret = 1;
+	    goto out2;
+	}
 
 	if (!date1)
-	    date1 = xmalloc (50);	/* plenty big :-) */
+	    date1 = xmalloc (MAXDATELEN);
 	*date1 = '\0';
 	if (RCS_getrevtime (rcsfile, vers_head, date1, 1) == -1)
 	{
 	    if (!really_quiet)
 		error (0, 0, "cannot find date in rcs file %s revision %s",
 		       rcs, vers_head);
-	    return (1);
+	    ret = 1;
+	    goto out2;
 	}
     }
     vers_tag = RCS_getversion (rcsfile, rev1, date1, force_tag_match,
@@ -403,31 +447,56 @@ patch_fileproc (callerdat, finfo)
     }
 
     if (vers_tag == NULL && vers_head == NULL)
-	return (0);			/* nothing known about specified revs */
+    {
+	/* Nothing known about specified revs.  */
+	ret = 0;
+	goto out2;
+    }
 
     if (vers_tag && vers_head && strcmp (vers_head, vers_tag) == 0)
-	return (0);			/* not changed between releases */
+    {
+	/* Not changed between releases.  */
+	ret = 0;
+	goto out2;
+    }
 
     if (patch_short)
     {
-	(void) printf ("File %s ", finfo->fullname);
+	cvs_output ("File ", 0);
+	cvs_output (finfo->fullname, 0);
 	if (vers_tag == NULL)
-	    (void) printf ("is new; current revision %s\n", vers_head);
+	{
+	    cvs_output (" is new; current revision ", 0);
+	    cvs_output (vers_head, 0);
+	    cvs_output ("\n", 1);
+	}
 	else if (vers_head == NULL)
 	{
-	    (void) printf ("is removed; not included in ");
+	    cvs_output (" is removed; not included in ", 0);
 	    if (rev2 != NULL)
-		(void) printf ("release tag %s", rev2);
+	    {
+		cvs_output ("release tag ", 0);
+		cvs_output (rev2, 0);
+	    }
 	    else if (date2 != NULL)
-		(void) printf ("release date %s", date2);
+	    {
+		cvs_output ("release date ", 0);
+		cvs_output (date2, 0);
+	    }
 	    else
-		(void) printf ("current release");
-	    (void) printf ("\n");
+		cvs_output ("current release", 0);
+	    cvs_output ("\n", 1);
 	}
 	else
-	    (void) printf ("changed from revision %s to %s\n",
-			   vers_tag, vers_head);
-	return (0);
+	{
+	    cvs_output (" changed from revision ", 0);
+	    cvs_output (vers_tag, 0);
+	    cvs_output (" to ", 0);
+	    cvs_output (vers_head, 0);
+	    cvs_output ("\n", 1);
+	}
+	ret = 0;
+	goto out2;
     }
     tmpfile1 = cvs_temp_name ();
     if ((fp1 = CVS_FOPEN (tmpfile1, "w+")) != NULL)
@@ -449,7 +518,8 @@ patch_fileproc (callerdat, finfo)
     if (vers_tag != NULL)
     {
 	retcode = RCS_checkout (rcsfile, (char *) NULL, vers_tag,
-				(char *) NULL, options, tmpfile1);
+				rev1, options, tmpfile1,
+				(RCSCHECKOUTPROC) NULL, (void *) NULL);
 	if (retcode != 0)
 	{
 	    if (!really_quiet)
@@ -461,7 +531,9 @@ patch_fileproc (callerdat, finfo)
 	memset ((char *) &t, 0, sizeof (t));
 	if ((t.actime = t.modtime = RCS_getrevtime (rcsfile, vers_tag,
 						    (char *) 0, 0)) != -1)
-		(void) utime (tmpfile1, &t);
+	    /* I believe this timestamp only affects the dates in our diffs,
+	       and therefore should be on the server, not the client.  */
+	    (void) utime (tmpfile1, &t);
     }
     else if (toptwo_diffs)
     {
@@ -471,7 +543,8 @@ patch_fileproc (callerdat, finfo)
     if (vers_head != NULL)
     {
 	retcode = RCS_checkout (rcsfile, (char *) NULL, vers_head,
-				(char *) NULL, options, tmpfile2);
+				rev2, options, tmpfile2,
+				(RCSCHECKOUTPROC) NULL, (void *) NULL);
 	if (retcode != 0)
 	{
 	    if (!really_quiet)
@@ -482,18 +555,12 @@ patch_fileproc (callerdat, finfo)
 	}
 	if ((t.actime = t.modtime = RCS_getrevtime (rcsfile, vers_head,
 						    (char *) 0, 0)) != -1)
-		(void) utime (tmpfile2, &t);
+	    /* I believe this timestamp only affects the dates in our diffs,
+	       and therefore should be on the server, not the client.  */
+	    (void) utime (tmpfile2, &t);
     }
-    run_setup ("%s -%c", DIFF, unidiff ? 'u' : 'c');
-    run_arg (tmpfile1);
-    run_arg (tmpfile2);
 
-    line1 = NULL;
-    line1_chars_allocated = 0;
-    line2 = NULL;
-    line2_chars_allocated = 0;
-
-    switch (run_exec (RUN_TTY, tmpfile3, RUN_TTY, RUN_REALLY))
+    switch (diff_exec (tmpfile1, tmpfile2, unidiff ? "-u" : "-c", tmpfile3))
     {
 	case -1:			/* fork/wait failure */
 	    error (1, errno, "fork for diff failed on %s", rcs);
@@ -508,9 +575,9 @@ patch_fileproc (callerdat, finfo)
 	     */
 
 	    /* Output an "Index:" line for patch to use */
-	    (void) fflush (stdout);
-	    (void) printf ("Index: %s\n", finfo->fullname);
-	    (void) fflush (stdout);
+	    cvs_output ("Index: ", 0);
+	    cvs_output (finfo->fullname, 0);
+	    cvs_output ("\n", 1);
 
 	    fp = open_file (tmpfile3, "r");
 	    if (getline (&line1, &line1_chars_allocated, fp) < 0 ||
@@ -549,41 +616,72 @@ patch_fileproc (callerdat, finfo)
 		}
 	    }
 	    if (CVSroot_directory != NULL)
+	    {
+		strippath = xmalloc (strlen (CVSroot_directory) + 10);
 		(void) sprintf (strippath, "%s/", CVSroot_directory);
+	    }
 	    else
-		(void) strcpy (strippath, REPOS_STRIP);
+		strippath = xstrdup (REPOS_STRIP);
 	    if (strncmp (rcs, strippath, strlen (strippath)) == 0)
 		rcs += strlen (strippath);
+	    free (strippath);
 	    if (vers_tag != NULL)
 	    {
+		file1 = xmalloc (strlen (finfo->fullname)
+				 + strlen (vers_tag)
+				 + 10);
 		(void) sprintf (file1, "%s:%s", finfo->fullname, vers_tag);
 	    }
 	    else
 	    {
-		(void) strcpy (file1, DEVNULL);
+		file1 = xstrdup (DEVNULL);
 	    }
+	    file2 = xmalloc (strlen (finfo->fullname)
+			     + (vers_head != NULL ? strlen (vers_head) : 10)
+			     + 10);
 	    (void) sprintf (file2, "%s:%s", finfo->fullname,
 			    vers_head ? vers_head : "removed");
 
-	    /* Note that this prints "diff" not DIFF.  The format of a diff
-	       does not depend on the name of the program which happens to
-	       have produced it.  */
+	    /* Note that the string "diff" is specified by POSIX (for -c)
+	       and is part of the diff output format, not the name of a
+	       program.  */
 	    if (unidiff)
 	    {
-		(void) printf ("diff -u %s %s\n", file1, file2);
-		(void) printf ("--- %s%s+++ ", file1, cp1);
+		cvs_output ("diff -u ", 0);
+		cvs_output (file1, 0);
+		cvs_output (" ", 1);
+		cvs_output (file2, 0);
+		cvs_output ("\n", 1);
+
+		cvs_output ("--- ", 0);
+		cvs_output (file1, 0);
+		cvs_output (cp1, 0);
+		cvs_output ("+++ ", 0);
 	    }
 	    else
 	    {
-		(void) printf ("diff -c %s %s\n", file1, file2);
-		(void) printf ("*** %s%s--- ", file1, cp1);
+		cvs_output ("diff -c ", 0);
+		cvs_output (file1, 0);
+		cvs_output (" ", 1);
+		cvs_output (file2, 0);
+		cvs_output ("\n", 1);
+
+		cvs_output ("*** ", 0);
+		cvs_output (file1, 0);
+		cvs_output (cp1, 0);
+		cvs_output ("--- ", 0);
 	    }
 
-	    (void) printf ("%s%s", finfo->fullname, cp2);
+	    cvs_output (finfo->fullname, 0);
+	    cvs_output (cp2, 0);
+
 	    /* spew the rest of the diff out */
 	    while (getline (&line1, &line1_chars_allocated, fp) >= 0)
-		(void) fputs (line1, stdout);
+		cvs_output (line1, 0);
+
 	    (void) fclose (fp);
+	    free (file1);
+	    free (file2);
 	    break;
 	default:
 	    error (0, 0, "diff failed for %s", finfo->fullname);
@@ -601,6 +699,10 @@ patch_fileproc (callerdat, finfo)
     free (tmpfile2);
     free (tmpfile3);
     tmpfile1 = tmpfile2 = tmpfile3 = NULL;
+
+ out2:
+    if (rcs != NULL)
+	free (rcs);
     return (ret);
 }
 
