@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Rtag
  * 
@@ -72,8 +72,10 @@ static const char *const rtag_usage[] =
     "\t-n\tNo execution of 'tag program'\n",
     "\t-d\tDelete the given Tag.\n",
     "\t-b\tMake the tag a \"branch\" tag, allowing concurrent development.\n",
-    "\t-[rD]\tExisting tag or Date.\n",
+    "\t-r rev\tExisting revision/tag.\n",
+    "\t-D\tExisting date.\n",
     "\t-F\tMove tag if it already exists\n",	
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
@@ -91,8 +93,8 @@ rtag (argc, argv)
     if (argc == -1)
 	usage (rtag_usage);
 
-    optind = 1;
-    while ((c = getopt (argc, argv, "FanfQqlRdbr:D:")) != -1)
+    optind = 0;
+    while ((c = getopt (argc, argv, "+FanfQqlRdbr:D:")) != -1)
     {
 	switch (c)
 	{
@@ -164,9 +166,11 @@ rtag (argc, argv)
     {
 	/* We're the client side.  Fire up the remote server.  */
 	start_server ();
-	
+
 	ign_setup ();
 
+	if (!force_tag_match)
+	    send_arg ("-f");
 	if (local)
 	    send_arg("-l");
 	if (delete_flag)
@@ -175,7 +179,7 @@ rtag (argc, argv)
 	    send_arg("-b");
 	if (force_tag_move)
 	    send_arg("-F");
-	if (run_module_prog)
+	if (!run_module_prog)
 	    send_arg("-n");
 	if (attic_too)
 	    send_arg("-a");
@@ -204,7 +208,8 @@ rtag (argc, argv)
 	/* XXX last arg should be repository, but doesn't make sense here */
 	history_write ('T', (delete_flag ? "D" : (numtag ? numtag : 
 		       (date ? date : "A"))), symtag, argv[i], "");
-	err += do_module (db, argv[i], TAG, delete_flag ? "Untagging" : "Tagging",
+	err += do_module (db, argv[i], TAG,
+			  delete_flag ? "Untagging" : "Tagging",
 			  rtag_proc, (char *) NULL, 0, 0, run_module_prog,
 			  symtag);
     }
@@ -229,19 +234,25 @@ rtag_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
     char *mname;
     char *msg;
 {
+    /* Begin section which is identical to patch_proc--should this
+       be abstracted out somehow?  */
     int err = 0;
     int which;
-    char repository[PATH_MAX];
-    char where[PATH_MAX];
+    char *repository;
+    char *where;
 
+    repository = xmalloc (strlen (CVSroot_directory) + strlen (argv[0])
+			  + (mfile == NULL ? 0 : strlen (mfile)) + 30);
     (void) sprintf (repository, "%s/%s", CVSroot_directory, argv[0]);
+    where = xmalloc (strlen (argv[0]) + (mfile == NULL ? 0 : strlen (mfile))
+		     + 10);
     (void) strcpy (where, argv[0]);
 
     /* if mfile isn't null, we need to set up to do only part of the module */
     if (mfile != NULL)
     {
 	char *cp;
-	char path[PATH_MAX];
+	char *path;
 
 	/* if the portion of the module is a path, put the dir part on repos */
 	if ((cp = strrchr (mfile, '/')) != NULL)
@@ -255,6 +266,7 @@ rtag_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 	}
 
 	/* take care of the rest */
+	path = xmalloc (strlen (repository) + strlen (mfile) + 5);
 	(void) sprintf (path, "%s/%s", repository, mfile);
 	if (isdir (path))
 	{
@@ -273,14 +285,18 @@ rtag_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 	    argv[1] = xstrdup (mfile);
 	    (*pargc) = 2;
 	}
+	free (path);
     }
 
-    /* chdir to the starting directory */
+    /* cd to the starting repository */
     if ( CVS_CHDIR (repository) < 0)
     {
 	error (0, errno, "cannot chdir to %s", repository);
+	free (repository);
 	return (1);
     }
+    free (repository);
+    /* End section which is identical to patch_proc.  */
 
     if (delete_flag || attic_too || (force_tag_match && numtag))
 	which = W_REPOS | W_ATTIC;
@@ -312,7 +328,7 @@ rtag_proc (pargc, argv, xwhere, mwhere, mfile, shorten, local_specified,
 			   (DIRLEAVEPROC) NULL, NULL,
 			   *pargc - 1, argv + 1, local,
 			   which, 0, 0, where, 1);
-
+    free (where);
     dellist(&mtlist);
 
     return (err);
@@ -456,11 +472,10 @@ pretag_proc(repository, filter)
         }
         free(s);
     }
-    run_setup("%s %s %s %s",
-              filter,
-              symtag,
-              delete_flag ? "del" : force_tag_move ? "mov" : "add",
-              repository);
+    run_setup (filter);
+    run_arg (symtag);
+    run_arg (delete_flag ? "del" : force_tag_move ? "mov" : "add");
+    run_arg (repository);
     walklist(tlist, pretag_list_proc, NULL);
     return (run_exec(RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL|RUN_REALLY));
 }
@@ -518,7 +533,15 @@ rtag_fileproc (callerdat, finfo)
 
     /* Lock the directory if it is not already locked.  We might be
        able to rely on rtag_dirproc for this.  */
-    tag_lockdir (finfo->repository);
+
+    /* It would be nice to provide consistency with respect to
+       commits; however CVS lacks the infrastructure to do that (see
+       Concurrency in cvs.texinfo and comment in do_recursion).  We
+       can and will prevent simultaneous tag operations from
+       interfering with each other, by write locking each directory as
+       we enter it, and unlocking it as we leave it.  */
+
+    lock_dir_for_write (finfo->repository);
 
     /* find the parsed RCS data */
     if ((rcsfile = finfo->rcs) == NULL)
@@ -573,6 +596,8 @@ rtag_fileproc (callerdat, finfo)
 	 */
 	rev = branch_mode ? RCS_magicrev (rcsfile, version) : numtag;
 	retcode = RCS_settag(rcsfile, symtag, numtag);
+	if (retcode == 0)
+	    RCS_rewrite (rcsfile, NULL, NULL);
     }
     else
     {
@@ -592,7 +617,7 @@ rtag_fileproc (callerdat, finfo)
 				   (int *) NULL);
 	if (oversion != NULL)
 	{
-	    int isbranch = RCS_isbranch (finfo->rcs, symtag);
+	    int isbranch = RCS_nodeisbranch (finfo->rcs, symtag);
 
 	    /*
 	     * if versions the same and neither old or new are branches don't
@@ -622,6 +647,8 @@ rtag_fileproc (callerdat, finfo)
 	    free (oversion);
 	}
 	retcode = RCS_settag(rcsfile, symtag, rev);
+	if (retcode == 0)
+	    RCS_rewrite (rcsfile, NULL, NULL);
     }
 
     if (retcode != 0)
@@ -673,7 +700,7 @@ rtag_delete (rcsfile)
 	return (0);
     free (version);
 
-    if ((retcode = RCS_deltag(rcsfile, symtag, 1)) != 0)
+    if ((retcode = RCS_deltag(rcsfile, symtag)) != 0)
     {
 	if (!quiet)
 	    error (0, retcode == -1 ? errno : 0,
@@ -681,6 +708,7 @@ rtag_delete (rcsfile)
 		   rcsfile->path);
 	return (1);
     }
+    RCS_rewrite (rcsfile, NULL, NULL);
     return (0);
 }
 
@@ -694,7 +722,7 @@ rtag_filesdoneproc (callerdat, err, repos, update_dir, entries)
     char *update_dir;
     List *entries;
 {
-    tag_unlockdir ();
+    Lock_Cleanup ();
 
     return (err);
 }
@@ -712,7 +740,8 @@ rtag_dirproc (callerdat, dir, repos, update_dir, entries)
     List *entries;
 {
     if (!quiet)
-	error (0, 0, "%s %s", delete_flag ? "Untagging" : "Tagging", update_dir);
+	error (0, 0, "%s %s", delete_flag ? "Untagging" : "Tagging",
+	       update_dir);
     return (R_PROCESS);
 }
 
