@@ -10,11 +10,7 @@
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   GNU General Public License for more details.  */
 
 #include "cvs.h"
 
@@ -113,50 +109,32 @@ run_arg (s)
     run_add_arg (s);
 }
 
-/* VARARGS */
-#if defined (HAVE_VPRINTF) && (defined (USE_PROTOTYPES) ? USE_PROTOTYPES : defined (__STDC__))
-void 
-run_args (const char *fmt,...)
-#else
-void 
-run_args (fmt, va_alist)
-    char *fmt;
-    va_dcl
-#endif
-{
-#ifdef HAVE_VPRINTF
-    va_list args;
-#endif
-
-    run_init_prog ();
-
-    /* process the varargs into run_prog */
-#ifdef HAVE_VPRINTF
-    VA_START (args, fmt);
-    (void) vsprintf (run_prog, fmt, args);
-    va_end (args);
-#else
-    (void) sprintf (run_prog, fmt, a1, a2, a3, a4, a5, a6, a7, a8);
-#endif
-
-    /* and add the (single) argument to the run_argv list */
-    run_add_arg (run_prog);
-}
-
 /* Return a malloc'd copy of s, with double quotes around it.  */
 static char *
 quote (const char *s)
 {
-    size_t s_len = strlen (s);
-    char *copy = xmalloc (s_len + 3);
-    char *scan = copy;
+    size_t s_len = 0;
+    char *copy = NULL;
+    char *scan = (char *) s;
 
+    /* scan string for extra quotes ... */
+    while (*scan)
+	if ('"' == *scan++)
+	    s_len += 2;   /* one extra for the quote character */
+	else
+	    s_len++;
+    /* allocate length + byte for ending zero + for double quotes around */
+    scan = copy = xmalloc(s_len + 3);
     *scan++ = '"';
-    strcpy (scan, s);
-    scan += s_len;
+    while (*s)
+    {
+	if ('"' == *s)
+	    *scan++ = '\\';
+	*scan++ = *s++;
+    }
+    /* ending quote and closing zero */
     *scan++ = '"';
     *scan++ = '\0';
-
     return copy;
 }
 
@@ -343,7 +321,6 @@ run_exec (stin, stout, sterr, flags)
     return (status);
 }
 
-
 void
 run_print (fp)
     FILE *fp;
@@ -392,7 +369,22 @@ run_popen (cmd, mode)
        double quotes.  */
     {
         char *requoted = requote (cmd);
+	/* Save and restore our file descriptors to work around
+	   apparent bugs in _popen.  We are perhaps better off using
+	   the win32 functions instead of _popen.  */
+	int old_stdin = dup (STDIN_FILENO);
+	int old_stdout = dup (STDOUT_FILENO);
+	int old_stderr = dup (STDERR_FILENO);
+
 	FILE *result = popen (requoted, mode);
+
+	dup2 (old_stdin, STDIN_FILENO);
+	dup2 (old_stdout, STDOUT_FILENO);
+	dup2 (old_stderr, STDERR_FILENO);
+	close (old_stdin);
+	close (old_stdout);
+	close (old_stderr);
+
 	free (requoted);
 	return result;
     }
